import functools

from PyQt5.QtCore import Qt
from PyQt5.QtCore import QCoreApplication

from PyQt5.QtWidgets import QTreeView
from PyQt5.QtWidgets import QCheckBox
from PyQt5.QtWidgets import QTableWidgetItem
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QHeaderView
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QAbstractItemView

from PyQt5.QtGui import QStandardItemModel
from PyQt5.QtGui import QBrush
from PyQt5.QtGui import QColor

from galacteek import ensure
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs import ipfsOp
from galacteek.core.modelhelpers import UneditableItem
from galacteek.core.modelhelpers import modelSearch

from .widgets import GalacteekTab

from .ui_batchpinlist import Ui_BatchPinList
from .helpers import messageBox

from .i18n import iPath
from .i18n import iCidOrPath
from .i18n import iUnknown
from .i18n import iPinned
from .i18n import iPinning
from .i18n import iPin
from .i18n import iDoNotPin
from .i18n import iPinSingle
from .i18n import iPinRecursive
from .i18n import iCancel
from .i18n import iInvalidInput


def iQueue():
    return QCoreApplication.translate('PinStatusWidget', 'Queue')


def iStatus():
    return QCoreApplication.translate('PinStatusWidget', 'Status')


def iNodesProcessed():
    return QCoreApplication.translate('PinStatusWidget', 'Nodes')


class PinStatusWidget(GalacteekTab):
    COL_TS = 0
    COL_QUEUE = 1
    COL_PATH = 2
    COL_STATUS = 3
    COL_PROGRESS = 4
    COL_CTRL = 5

    def __init__(self, gWindow):
        super(PinStatusWidget, self).__init__(gWindow)

        self.tree = QTreeView()
        self.tree.setObjectName('pinStatusWidget')
        self.boxLayout = QVBoxLayout()
        self.boxLayout.addWidget(self.tree)

        self.ctrlLayout = QHBoxLayout()
        self.btnPin = QPushButton(iPin())
        self.pathLabel = QLabel(iCidOrPath())
        self.pathEdit = QLineEdit()
        self.ctrlLayout.addWidget(self.pathLabel)
        self.ctrlLayout.addWidget(self.pathEdit)
        self.ctrlLayout.addWidget(self.btnPin)
        self.vLayout.addLayout(self.ctrlLayout)
        self.vLayout.addLayout(self.boxLayout)

        self.app.ipfsCtx.pinItemStatusChanged.connect(self.onPinStatusChanged)
        self.app.ipfsCtx.pinFinished.connect(self.onPinFinished)
        self.app.ipfsCtx.pinItemRemoved.connect(self.onItemRemoved)
        self.pathEdit.returnPressed.connect(self.onPathEntered)
        self.btnPin.clicked.connect(self.onPathEntered)

        self.model = QStandardItemModel()
        self.model.setHorizontalHeaderLabels(
            ['TS', iQueue(), iPath(), iStatus(), iNodesProcessed(), ''])

        self.tree.setSortingEnabled(True)
        self.tree.setModel(self.model)
        self.tree.sortByColumn(self.COL_TS, Qt.DescendingOrder)

        for col in [self.COL_QUEUE, self.COL_PATH, self.COL_PROGRESS]:
            self.tree.header().setSectionResizeMode(
                col, QHeaderView.ResizeToContents)

        self.tree.hideColumn(self.COL_TS)

    def resort(self):
        self.model.sort(self.COL_TS, Qt.DescendingOrder)

    def onPathEntered(self):
        text = self.pathEdit.text()
        self.pathEdit.clear()

        path = IPFSPath(text)
        if path.valid:
            ensure(self.app.ipfsCtx.pinner.queue(
                path.objPath, True, None))
        else:
            messageBox(iInvalidInput())

    def removeItem(self, path):
        modelSearch(self.model,
                    search=path, columns=[self.COL_PATH],
                    delete=True)

    def onItemRemoved(self, qname, path):
        self.removeItem(path)

    def findPinItems(self, path):
        ret = modelSearch(self.model,
                          search=path, columns=[self.COL_PATH])
        if len(ret) == 0:
            return None

        itemP = self.model.itemFromIndex(ret.pop())

        if not itemP:
            return None

        idxQueue = self.model.index(itemP.row(), self.COL_QUEUE,
                                    itemP.index().parent())
        idxProgress = self.model.index(itemP.row(), self.COL_PROGRESS,
                                       itemP.index().parent())
        idxStatus = self.model.index(itemP.row(), self.COL_STATUS,
                                     itemP.index().parent())
        idxC = self.model.index(itemP.row(), self.COL_CTRL,
                                itemP.index().parent())
        cancelButton = self.tree.indexWidget(idxC)

        return {
            'itemPath': itemP,
            'itemQname': self.model.itemFromIndex(idxQueue),
            'itemProgress': self.model.itemFromIndex(idxProgress),
            'itemStatus': self.model.itemFromIndex(idxStatus),
            'cancelButton': cancelButton
        }

    def onPinFinished(self, path):
        items = self.findPinItems(path)

        if items:
            items['itemStatus'].setText(iPinned())
            items['itemProgress'].setText('OK')

            color = QBrush(QColor('#c1f0c1'))
            for item in [items['itemQname'],
                         items['itemPath'], items['itemStatus'],
                         items['itemProgress']]:
                item.setBackground(color)

            if items['cancelButton']:
                items['cancelButton'].setEnabled(False)

        self.resort()
        self.purgeFinishedItems()

    def purgeFinishedItems(self):
        maxFinished = 16
        ret = modelSearch(self.model,
                          search=iPinned(),
                          columns=[self.COL_STATUS])

        if len(ret) > maxFinished:
            rows = []
            for idx in ret:
                item = self.model.itemFromIndex(idx)
                if not item:
                    continue
                rows.append(item.row())

            try:
                for row in list(sorted(rows))[int(maxFinished / 2):]:
                    self.model.removeRow(row)
            except:
                pass

    def onCancel(self, qname, path):
        self.removeItem(path)
        self.app.ipfsCtx.pinner.cancel(qname, path)

    def onPinStatusChanged(self, qname, path, statusInfo):
        nodesProcessed = statusInfo['status'].get('Progress', iUnknown())
        items = self.findPinItems(path)

        if not items:
            btnCancel = QToolButton()
            btnCancel.setText(iCancel())
            btnCancel.clicked.connect(
                functools.partial(self.onCancel, qname, path))
            btnCancel.setFixedWidth(140)

            itemTs = UneditableItem(str(statusInfo['ts_queued']))
            itemQ = UneditableItem(qname)
            itemP = UneditableItem(path)
            itemP.setToolTip(path)
            itemStatus = UneditableItem(iPinning())
            itemNodes = UneditableItem(str(nodesProcessed))
            itemC = UneditableItem('')

            self.model.invisibleRootItem().appendRow(
                [itemTs, itemQ, itemP, itemStatus, itemNodes, itemC])
            idx = self.model.indexFromItem(itemC)
            self.tree.setIndexWidget(idx, btnCancel)
        else:
            items['itemProgress'].setText(str(nodesProcessed))

        self.resort()


class PinBatchTab(GalacteekTab):
    pass


class PinBatchWidget(QWidget):
    COL_PATH = 0
    COL_NOPIN = 1
    COL_SPIN = 2
    COL_RPIN = 3

    def __init__(self, basePath, pathList, parent=None):
        super(PinBatchWidget, self).__init__(parent)

        self.basePath = basePath
        self.pathList = pathList
        self._pinTask = None

        self.ui = Ui_BatchPinList()
        self.ui.setupUi(self)

        self.ui.labelBasePath.setText(
            'Base IPFS path: <b>{p}</b>'.format(p=str(self.basePath)))

        horizHeader = self.ui.tableWidget.horizontalHeader()
        horizHeader.sectionClicked.connect(self.onHorizSectionClicked)
        horizHeader.setSectionResizeMode(QHeaderView.ResizeToContents)
        self.ui.tableWidget.setHorizontalHeaderLabels(
            [iPath(), iDoNotPin(), iPinSingle(), iPinRecursive()])

        self.ui.tableWidget.verticalHeader().hide()
        self.ui.tableWidget.setRowCount(len(self.pathList))

        self.ui.proceedButton.clicked.connect(self.onPinObjects)
        self.ui.cancelButton.clicked.connect(self.onCancel)
        self.ui.cancelButton.hide()
        self.insertItems()

    def insertItems(self):
        added = []
        for path in self.pathList:
            if path.objPath in added:
                continue

            curRow = len(added)

            if len(path.objPath) > 92:
                text = path.objPath[0:92] + '...'
            else:
                text = path.objPath

            pItem = QTableWidgetItem(text)

            pItem.setData(Qt.UserRole, path.objPath)
            pItem.setToolTip(path.objPath)
            pItem.setFlags(
                Qt.ItemIsSelectable | Qt.ItemIsEnabled)

            checkBoxN = QCheckBox(self)
            checkBoxS = QCheckBox(self)
            checkBoxR = QCheckBox(self)

            self.ui.tableWidget.setItem(curRow, self.COL_PATH, pItem)
            self.ui.tableWidget.setCellWidget(
                curRow, self.COL_NOPIN, checkBoxN)
            self.ui.tableWidget.setCellWidget(curRow, self.COL_SPIN, checkBoxS)
            self.ui.tableWidget.setCellWidget(curRow, self.COL_RPIN, checkBoxR)

            checkBoxN.stateChanged.connect(
                functools.partial(self.noPinChecked, pItem.row()))
            checkBoxS.stateChanged.connect(
                functools.partial(self.pinSingleChecked, pItem.row()))
            checkBoxR.stateChanged.connect(
                functools.partial(self.pinRecursiveChecked, pItem.row()))
            checkBoxS.setCheckState(Qt.Checked)
            added.append(path.objPath)

        self.ui.tableWidget.setRowCount(len(added))

    def onHorizSectionClicked(self, section):
        if section in [self.COL_NOPIN, self.COL_SPIN, self.COL_RPIN]:
            for row in range(self.ui.tableWidget.rowCount()):
                item = self.ui.tableWidget.cellWidget(row, section)
                if item:
                    item.setCheckState(Qt.Checked)

    def disableAdjacent(self, row, col, state):
        adjacent = self.ui.tableWidget.cellWidget(row, col)
        if adjacent and state == Qt.Checked:
            adjacent.setCheckState(Qt.Unchecked)

    def noPinChecked(self, row, state):
        self.disableAdjacent(row, self.COL_SPIN, state)
        self.disableAdjacent(row, self.COL_RPIN, state)

    def pinSingleChecked(self, row, state):
        self.disableAdjacent(row, self.COL_NOPIN, state)
        self.disableAdjacent(row, self.COL_RPIN, state)

    def pinRecursiveChecked(self, row, state):
        self.disableAdjacent(row, self.COL_NOPIN, state)
        self.disableAdjacent(row, self.COL_SPIN, state)

    def onPinObjects(self):
        self.ui.proceedButton.setText(iPinning() + ' ...')
        self.ui.proceedButton.setEnabled(False)
        self._pinTask = ensure(self.pinSelectedObjects(),
                               futcallback=self.onPinFinished)
        self.ui.cancelButton.show()

    def onPinFinished(self, future):
        try:
            self.ui.cancelButton.hide()
            self.ui.proceedButton.setText('Done')
        except:
            pass

    def onCancel(self):
        if self._pinTask:
            self._pinTask.cancel()

        self.parentWidget().close()

    @ipfsOp
    async def pinSelectedObjects(self, ipfsop):
        def disableRow(pItem, *boxes):
            pItem.setFlags(Qt.NoItemFlags)
            [box.setEnabled(False) for box in boxes]

        for row in range(self.ui.tableWidget.rowCount()):
            pItem = self.ui.tableWidget.item(row, self.COL_PATH)
            noPinBox = self.ui.tableWidget.cellWidget(row, self.COL_NOPIN)
            pSingleBox = self.ui.tableWidget.cellWidget(row, self.COL_SPIN)
            pRecBox = self.ui.tableWidget.cellWidget(row, self.COL_RPIN)

            self.ui.tableWidget.scrollToItem(
                pItem, QAbstractItemView.PositionAtCenter)

            path = pItem.data(Qt.UserRole)
            if not path or noPinBox.isChecked():
                await ipfsop.sleep(0.3)
                disableRow(pItem, noPinBox, pSingleBox, pRecBox)
                continue

            if pSingleBox.isChecked() or pRecBox.isChecked():
                await ipfsop.ctx.pin(
                    path, recursive=pRecBox.isChecked(),
                    qname='browser-batch')
                disableRow(pItem, noPinBox, pSingleBox, pRecBox)

            await ipfsop.sleep(0.3)



import re
import uuid
import aioipfs
import functools
import asyncio
import aiodns
import async_timeout
import collections
import time
from datetime import datetime

from PyQt5.QtWebEngineCore import QWebEngineUrlSchemeHandler
from PyQt5.QtWebEngineCore import QWebEngineUrlScheme
from PyQt5.QtWebEngineCore import QWebEngineUrlRequestJob

from PyQt5.QtCore import QUrl
from PyQt5.QtCore import QIODevice
from PyQt5.QtCore import QBuffer
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QMutex

from galacteek import log
from galacteek import logUser
from galacteek import ensure
from galacteek.ipfs import ipfsOp
from galacteek.ipfs.mimetype import detectMimeTypeFromBuffer
from galacteek.ipfs.cidhelpers import joinIpfs
from galacteek.ipfs.cidhelpers import joinIpns
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.cidhelpers import ipfsRegSearchCid32
from galacteek.ipfs.cidhelpers import domainValid
from galacteek.ipfs.ipfsops import APIErrorDecoder
from galacteek.dweb.render import renderTemplate
from galacteek.dweb.enswhois import ensContentHash

from galacteek.ipdapps import dappsRegisterSchemes


# Core schemes
SCHEME_DWEB = 'dweb'
SCHEME_FS = 'fs'
SCHEME_IPFS = 'ipfs'
SCHEME_IPNS = 'ipns'

# ENS-related schemes (ensr is redirect-on-resolve ENS scheme)
SCHEME_ENS = 'ens'
SCHEME_ENSR = 'ensr'
SCHEME_E = 'e'

# Misc schemes
SCHEME_Z = 'z'
SCHEME_Q = 'q'
SCHEME_GALACTEEK = 'glk'
SCHEME_PALACE = 'palace'
SCHEME_MANUAL = 'manual'


# Default flags used by declareUrlScheme()
defaultSchemeFlags = QWebEngineUrlScheme.SecureScheme | \
    QWebEngineUrlScheme.ViewSourceAllowed

# Registered URL schemes
urlSchemes = {}


def allUrlSchemes():
    global urlSchemes
    return urlSchemes


def isSchemeRegistered(scheme):
    global urlSchemes
    for section, schemes in urlSchemes.items():
        if scheme in schemes:
            return True


def declareUrlScheme(name,
                     flags=defaultSchemeFlags,
                     syntax=QWebEngineUrlScheme.Syntax.Host,
                     defaultPort=QWebEngineUrlScheme.PortUnspecified,
                     schemeSection='core'):
    global urlSchemes
    scheme = QWebEngineUrlScheme(name.encode())
    scheme.setFlags(flags)
    scheme.setSyntax(syntax)
    scheme.setDefaultPort(defaultPort)
    QWebEngineUrlScheme.registerScheme(scheme)
    urlSchemes.setdefault(schemeSection, {})
    urlSchemes[schemeSection][name] = scheme
    return scheme


def registerMiscSchemes():
    declareUrlScheme(SCHEME_MANUAL,
                     syntax=QWebEngineUrlScheme.Syntax.Path,
                     schemeSection='misc'
                     )
    declareUrlScheme(SCHEME_Z,
                     syntax=QWebEngineUrlScheme.Syntax.Path,
                     flags=QWebEngineUrlScheme.LocalScheme,
                     schemeSection='misc'
                     )
    declareUrlScheme(SCHEME_Q,
                     syntax=QWebEngineUrlScheme.Syntax.Host,
                     flags=QWebEngineUrlScheme.LocalScheme,
                     schemeSection='core'
                     )


def initializeSchemes():
    name = QWebEngineUrlScheme.schemeByName(SCHEME_DWEB.encode()).name()
    if name:
        log.debug('initializeSchemes() already called')
        return

    declareUrlScheme(
        SCHEME_DWEB,
        syntax=QWebEngineUrlScheme.Syntax.Path,
    )

    declareUrlScheme(
        SCHEME_FS,
        syntax=QWebEngineUrlScheme.Syntax.Path,
    )

    declareUrlScheme(
        SCHEME_IPFS,
        syntax=QWebEngineUrlScheme.Syntax.Host
    )

    declareUrlScheme(
        SCHEME_IPNS,
        syntax=QWebEngineUrlScheme.Syntax.Host
    )

    declareUrlScheme(
        SCHEME_ENS,
        syntax=QWebEngineUrlScheme.Syntax.Host
    )

    declareUrlScheme(
        SCHEME_ENSR,
        syntax=QWebEngineUrlScheme.Syntax.Host
    )

    dappsRegisterSchemes()
    registerMiscSchemes()


def isIpfsUrl(url):
    if url.isValid():
        return url.scheme() in [SCHEME_FS, SCHEME_DWEB,
                                SCHEME_IPFS, SCHEME_IPNS]


def isNativeIpfsUrl(url):
    if url.isValid():
        return url.scheme() in [SCHEME_IPFS, SCHEME_IPNS]


class BaseURLSchemeHandler(QWebEngineUrlSchemeHandler):
    webProfileNeeded = None

    def reqFailed(self, request):
        return request.fail(QWebEngineUrlRequestJob.RequestFailed)

    def urlInvalid(self, request):
        return request.fail(QWebEngineUrlRequestJob.UrlInvalid)

    def urlNotFound(self, request):
        return request.fail(QWebEngineUrlRequestJob.UrlNotFound)

    def aborted(self, request):
        return request.fail(QWebEngineUrlRequestJob.RequestAborted)


class IPFSObjectProxyScheme:
    """
    For schemes that proxy/map IPFS objects
    """

    async def urlProxiedPath(self, url):
        """
        Return the object mapped for `url`
        """
        raise Exception('Implement urlProxiedPath()')


class ENSWhoisSchemeHandler(BaseURLSchemeHandler):
    """
    Deprecated ENS resolver (that was using api.whoisens.org)
    """

    contentReady = pyqtSignal(str, QWebEngineUrlRequestJob, str, bytes)
    domainResolved = pyqtSignal(str, IPFSPath)

    def __init__(self, app, parent=None):
        super(ENSWhoisSchemeHandler, self).__init__(parent)

        self.app = app
        self.requests = {}

    def onRequestDestroyed(self, uid):
        if uid in self.requests:
            del self.requests[uid]

    async def handleRequest(self, request):
        rUrl = request.requestUrl()
        if not rUrl.isValid():
            return self.urlInvalid(request)

        domain = rUrl.host()
        uPath = rUrl.path()

        if not domain or len(domain) > 512:
            logUser.info('ENS: invalid domain request')
            return self.urlInvalid(request)

        logUser.info('ENS: resolving {0}'.format(domain))

        path = await ensContentHash(domain, sslverify=self.app.sslverify)
        if path and path.valid:
            self.domainResolved.emit(domain, path)
            sPath = path.child(uPath) if uPath else path
            logUser.info('ENS: {domain} resolved to {res}'.format(
                domain=domain, res=sPath.ipfsUrl))
            return request.redirect(QUrl(sPath.ipfsUrl))
        else:
            logUser.info('ENS: {domain} resolve failed'.format(
                domain=domain))
            request.fail(QWebEngineUrlRequestJob.UrlInvalid)

    def requestStarted(self, request):
        uid = str(uuid.uuid4())
        self.requests[uid] = request
        request.destroyed.connect(lambda: self.onRequestDestroyed(uid))
        ensure(self.handleRequest(self.requests[uid]))


class EthDNSResolver:
    def __init__(self, loop):
        self.dnsResolver = aiodns.DNSResolver(loop=loop)

    def debug(self, msg):
        log.debug('EthDNS resolver: {}'.format(msg))

    async def resolveEnsDomain(self, domain, timeout=15):
        """
        Resolve an ENS domain with EthDNS, and return the
        IPFS path for this domain

        :param str domain: Eth domain name to resolve (e.g mydomain.eth)
        :rtype: IPFSPath
        """

        # EthDNS uses the .link TLD
        domainWLink = '{domain}.link'.format(domain=domain)

        try:
            with async_timeout.timeout(timeout):
                result = await self.dnsResolver.query(domainWLink, 'TXT')

                if not result:
                    self.debug('Failed to resolve {}'.format(domain))
                    return None

                for entry in result:
                    # Grab the dnslink

                    match = re.search('^dnslink=(.*)$', entry.text)
                    if not match:
                        continue

                    return IPFSPath(match.group(1), autoCidConv=True)
        except asyncio.TimeoutError:
            self.debug('Timeout resolving domain {}'.format(domain))
        except Exception as err:
            self.debug('Error while resolving domain {0}: {1}'.format(
                domain, str(err)))


class EthDNSSchemeHandler(BaseURLSchemeHandler):
    """
    ENS scheme handler (resolves ENS domains through EthDNS with aiodns)

    Redirects to the resolved IPFS object
    """

    contentReady = pyqtSignal(str, QWebEngineUrlRequestJob, str, bytes)
    domainResolved = pyqtSignal(str, IPFSPath)

    def __init__(self, app, resolver=None, parent=None):
        super(EthDNSSchemeHandler, self).__init__(parent)

        self.app = app
        self.requests = {}
        self.ethResolver = resolver if resolver else \
            EthDNSResolver(self.app.loop)

    def debug(self, msg):
        log.debug('EthDNS: {}'.format(msg))

    def onRequestDestroyed(self, uid):
        if uid in self.requests:
            del self.requests[uid]

    async def handleRequest(self, request):
        rUrl = request.requestUrl()
        if not rUrl.isValid():
            return self.urlInvalid(request)

        domain = rUrl.host()
        uPath = rUrl.path()

        if not domain or len(domain) > 512 or not domainValid(domain):
            logUser.info('EthDNS: invalid domain request')
            return self.urlInvalid(request)

        logUser.info('EthDNS: resolving {0}'.format(domain))

        path = await self.ethResolver.resolveEnsDomain(domain)

        if path and path.valid:
            self.domainResolved.emit(domain, path)
            sPath = path.child(uPath) if uPath else path
            logUser.info('EthDNS: {domain} resolved to {res}'.format(
                domain=domain, res=sPath.ipfsUrl))
            return request.redirect(QUrl(sPath.ipfsUrl))
        else:
            logUser.info('EthDNS: {domain} resolve failed'.format(
                domain=domain))
            return self.urlNotFound(request)

    def requestStarted(self, request):
        uid = str(uuid.uuid4())
        self.requests[uid] = request
        request.destroyed.connect(lambda: self.onRequestDestroyed(uid))
        ensure(self.handleRequest(self.requests[uid]))


class NativeIPFSSchemeHandler(BaseURLSchemeHandler):
    """
    Native scheme handler for URLs using the ipfs://<cidv1base32>
    or ipns://<fqdn>/.. formats

    We don't use the gateway at all here, making the async requests
    manually on the daemon.

    The requestStarted() function will first assign a UUID to the
    request and store the request in the handler so that it doesn't get
    garbage-collected by QtWebEngine. The handleRequest() coroutine
    is responsible for fetching the objects from IPFS. When the data
    is ready it will call the renderData() coroutine which will detect
    the MIME type, and emit the 'contentReady' signal which is handled
    by the 'onContent' callback.
    """

    # contentReady signal (handled by onContent())
    contentReady = pyqtSignal(
        str, QWebEngineUrlRequestJob, IPFSPath, str, bytes)

    # objectServed signal: emitted when an object has been fetched and
    # served to a QtWebEngine request
    objectServed = pyqtSignal(IPFSPath, str, float)

    def __init__(self, app, parent=None, validCidQSize=32, reqTimeout=60 * 10,
                 noMutexes=False):
        super(NativeIPFSSchemeHandler, self).__init__(parent)

        self.app = app
        self.noMutexes = noMutexes
        self.requests = {}
        self.validCids = collections.deque([], validCidQSize)
        self.requestTimeout = reqTimeout
        self.contentReady.connect(self.onContent)

    def debug(self, msg):
        log.debug('Native scheme handler: {}'.format(msg))

    def onRequestDestroyed(self, uid):
        if uid in self.requests:
            del self.requests[uid]

    def onContent(self, uid, request, ipfsPath, ctype, data):
        if uid not in self.requests:
            # Destroyed ?
            return

        mutex = self.requests[uid]['mutex']

        if mutex and not self.noMutexes:
            mutex.lock()

        try:
            buf = self.requests[uid]['iodev']
            buf.open(QIODevice.WriteOnly)
            buf.write(data)
            buf.seek(0)
            buf.close()
            request.reply(ctype.encode('ascii'), buf)

            self.objectServed.emit(ipfsPath, ctype, time.time())

            if mutex and not self.noMutexes:
                mutex.unlock()
        except Exception:
            if mutex and not self.noMutexes:
                mutex.unlock()

            log.debug('Error buffering request data')
            request.fail(QWebEngineUrlRequestJob.RequestFailed)

    async def directoryListing(self, request, client, path):
        currentIpfsPath = IPFSPath(path)

        if not currentIpfsPath.valid:
            self.debug('Invalid path: {0}'.format(path))
            return

        ctx = {}
        try:
            listing = await client.core.ls(path)
        except aioipfs.APIError:
            self.debug('Error listing directory for path: {0}'.format(path))
            return None

        if 'Objects' not in listing:
            return None

        ctx['path'] = path
        ctx['links'] = []

        for obj in listing['Objects']:
            if not isinstance(obj, dict) or 'Hash' not in obj:
                continue

            if obj['Hash'] == path:
                ctx['links'] += obj.get('Links', [])

        for lnk in ctx['links']:
            lType = lnk.get('Type')

            if lType == 1:
                lnk['href'] = lnk['Name'] + '/'
            else:
                lnk['href'] = lnk['Name']

        try:
            data = await renderTemplate('ipfsdirlisting.html', **ctx)
        except:
            self.debug('Failed to render directory listing template')
        else:
            return data.encode()

    async def renderDirectory(self, request, client, path):
        ipfsPath = IPFSPath(path)
        indexPath = ipfsPath.child('index.html')

        try:
            data = await client.cat(str(indexPath))
        except aioipfs.APIError as exc:
            dec = APIErrorDecoder(exc)

            if dec.errNoSuchLink():
                return await self.directoryListing(request, client, path)

            return self.urlInvalid(request)
        else:
            return data

    async def renderDagListing(self, request, client, path, **kw):
        return b'ENOTIMPLEMENTED'

    async def renderDagNode(self, request, client, ipfsPath, **kw):
        self.debug('DAG node render: {}'.format(ipfsPath))
        indexPath = ipfsPath.child('index.html')

        try:
            stat = None
            try:
                with async_timeout.timeout(8):
                    stat = await client.object.stat(str(indexPath))
            except asyncio.TimeoutError:
                return self.urlInvalid(request)
            except Exception:
                return self.reqFailed(request)

            if not stat:
                return await self.renderDagListing(request, client, ipfsPath,
                                                   **kw)
            else:
                # The index is present, redirect
                urlR = QUrl(indexPath.ipfsUrl)

                if urlR.isValid():
                    request.redirect(urlR)
        except aioipfs.APIError:
            return await self.renderDagListing(request, client, ipfsPath, **kw)

    async def renderData(self, request, ipfsPath, data, uid):
        cType = await detectMimeTypeFromBuffer(data[0:512])

        if cType:
            self.contentReady.emit(uid, request, ipfsPath, cType.type, data)
        else:
            self.debug('Impossible to detect MIME type for URL: {0}'.format(
                request.requestUrl().toString()))
            self.contentReady.emit(
                uid, request, 'application/octet-stream', data)

    @ipfsOp
    async def handleRequest(self, ipfsop, request, uid):
        rUrl = request.requestUrl()
        scheme = rUrl.scheme()
        host = rUrl.host()

        if not host:
            return self.urlInvalid(request)

        if scheme == SCHEME_IPFS:
            # hostname = base32-encoded CID or FQDN
            #
            # We keep a list (deque) of CIDs that have been validated.
            # If you hit a page which references a lot of other resources
            # for instance, this should be faster than regexp

            if domainValid(host):
                ipfsPathS = joinIpns(host) + rUrl.path()
            else:
                if host not in self.validCids:
                    if not ipfsRegSearchCid32(host):
                        return self.urlInvalid(request)

                    self.validCids.append(host)

                ipfsPathS = joinIpfs(host) + rUrl.path()
        elif scheme == SCHEME_IPNS:
            ipfsPathS = joinIpns(host) + rUrl.path()
        else:
            return self.urlInvalid(request)

        ipfsPath = IPFSPath(ipfsPathS)
        if not ipfsPath.valid:
            return self.urlInvalid(request)

        try:
            with async_timeout.timeout(self.requestTimeout):
                return await self.fetchFromPath(ipfsop, request, ipfsPath, uid)
        except asyncio.TimeoutError:
            return self.reqFailed(request)
        except Exception:
            # Any other error
            return self.reqFailed(request)

    async def fetchFromPath(self, ipfsop, request, ipfsPath, uid, **kw):
        try:
            data = await ipfsop.client.cat(ipfsPath.objPath)
        except aioipfs.APIError as exc:
            await asyncio.sleep(0)

            self.debug('API error ({path}): {err}'.format(
                path=str(ipfsPath), err=exc.message)
            )

            dec = APIErrorDecoder(exc)

            if dec.errNoSuchLink():
                return self.urlNotFound(request)

            if dec.errIsDirectory():
                data = await self.renderDirectory(
                    request,
                    ipfsop.client,
                    str(ipfsPath)
                )
                if data:
                    return await self.renderData(request, ipfsPath, data, uid)

            if dec.errUnknownNode():
                # DAG / TODO
                data = await self.renderDagNode(
                    request,
                    ipfsop.client,
                    ipfsPath
                )
                if data:
                    return await self.renderData(request, ipfsPath, data, uid)
        else:
            return await self.renderData(request, ipfsPath, data, uid)

    def requestStarted(self, request):
        if not request.requestUrl().isValid():
            self.debug('Invalid URL: {}'.format(
                request.requestUrl().toString()))
            return self.urlInvalid(request)

        uid = str(uuid.uuid4())
        self.requests[uid] = {
            'request': request,
            'iodev': QBuffer(parent=request),
            'mutex': QMutex() if not self.noMutexes else None
        }

        request.destroyed.connect(
            functools.partial(self.onRequestDestroyed, uid))
        ensure(self.handleRequest(request, uid))


class ObjectProxySchemeHandler(NativeIPFSSchemeHandler, IPFSObjectProxyScheme):
    """
    This scheme handler acts as a "proxy" to an IPFS path.

    For example, by mapping the url scheme 'docs' to this path:

    /ipfs/bafybeihw6pfoai7wbyt5jhbiialicees3vn336fkrdmjlbheh2dqsxsmlu/

    Accessing 'docs:/pdf/today.pdf' will access the IPFS object at

    /ipfs/bafybeihw6pfoai7wbyt5jhbiialicees3vn336fkrdmjlbheh2dqsxsmlu/pdf/today.pdf
    """

    def __init__(self, app, ipfsPath, parent=None):
        """
        :param IPFSPath ipfsPath: the path to proxy to
        """
        super(ObjectProxySchemeHandler, self).__init__(app, parent=parent)

        self.proxiedPath = ipfsPath

    async def urlProxiedPath(self, url):
        return self.proxiedPath.child(url.path())

    @ipfsOp
    async def handleRequest(self, ipfsop, request, uid):
        if not self.proxiedPath.valid:
            return request.fail(QWebEngineUrlRequestJob.UrlInvalid)

        rUrl = request.requestUrl()
        ipfsPath = await self.urlProxiedPath(rUrl)

        if not ipfsPath or not ipfsPath.valid:
            return self.urlInvalid(request)

        return await self.fetchFromPath(ipfsop, request, ipfsPath, uid)


class MultiObjectHostSchemeHandler(NativeIPFSSchemeHandler,
                                   IPFSObjectProxyScheme):
    """
    Somehow similar to ObjectProxySchemeHandler, but this handler
    is host-based, the host being the name of a user-defined
    name-to-ipfs-path mapping. Used by the q:// scheme.
    A URL would look like this (with foo being the mapping name)::

        q://foo/src/luser.c

    If the mapped object is an IPNS path, it's periodically
    resolved and the result is cached.
    """

    def __init__(self, app, parent=None):
        super(MultiObjectHostSchemeHandler, self).__init__(app, parent=parent)

        self.app.towers['schemes'].qMappingsChanged.connect(
            self.onMappingsChanged)

        self.lock = asyncio.Lock()
        self.mappings = {}
        self._resolverTask = None

    async def urlProxiedPath(self, rUrl):
        host = rUrl.host()

        if host not in self.mappings:
            return None

        mapping = self.mappings[host]
        rCached = mapping['rcache']
        mappedTo = mapping['path']

        usedPath = rCached if rCached and rCached.valid else mappedTo

        if not usedPath.valid:
            return None

        path = rUrl.path()
        ipfsPath = usedPath.child(path)
        return ipfsPath

    def onMappingsChanged(self):
        ensure(self.updateMappings())

    async def updateMappings(self):
        mappings = self.app.marksLocal.qaGetMappings()
        for mapping in mappings:
            if mapping.name in self.mappings:
                continue

            with await self.lock:
                self.mappings[mapping.name] = {
                    'path': IPFSPath(mapping.path),
                    'rfrequency': mapping.ipnsFreq,
                    'rcache': None,
                    'rlast': None
                }

    async def start(self, **kw):
        if not self._resolverTask:
            self._resolverTask = ensure(self.mappingsResolver())

        await self.updateMappings()

    @ipfsOp
    async def mappingsResolver(self, ipfsop):
        while True:
            with await self.lock:
                for name, m in self.mappings.items():
                    now = datetime.now()
                    path = m['path']

                    if not path.isIpns:
                        continue

                    if not m['rlast'] or \
                            (now - m['rlast']).seconds > m['rfrequency']:
                        resolved = await path.resolve(ipfsop, noCache=True)
                        if resolved:
                            m['rcache'] = IPFSPath(resolved)
                            m['rlast'] = now

            await asyncio.sleep(60)

    @ipfsOp
    async def handleRequest(self, ipfsop, request, uid):
        rUrl = request.requestUrl()

        if not rUrl.host():
            return self.urlInvalid(request)

        ipfsPath = await self.urlProxiedPath(rUrl)

        if not ipfsPath or not ipfsPath.valid:
            return self.urlInvalid(request)

        return await self.fetchFromPath(ipfsop, request, ipfsPath, uid)


class DAGProxySchemeHandler(NativeIPFSSchemeHandler):
    """
    This scheme handler acts as a "proxy" to a DAG
    """

    def __init__(self, app, parent=None):
        super(DAGProxySchemeHandler, self).__init__(app, parent=parent)

        # The DAG being proxied
        self.proxied = None

    async def getDag(self):
        raise Exception('implement getDag()')

    async def renderDagListing(self, request, client, path):
        listing = await self.proxied.list(
            path=path.subPath if path.subPath else '')
        if not listing:
            return b'ELISTING'

        ctx = {}
        ctx['path'] = str(path)
        ctx['links'] = []

        for nname in listing:
            child = path.child(nname)
            ctx['links'].append({
                'name': nname,
                'href': child.ipfsUrl
            })

        try:
            data = await renderTemplate('ipfsdagnodelisting.html', **ctx)
        except:
            pass
        else:
            return data.encode()

    @ipfsOp
    async def handleRequest(self, ipfsop, request, uid):
        if self.proxied is None:
            self.proxied = await self.getDag()

        if not self.proxied:
            return request.fail(QWebEngineUrlRequestJob.UrlInvalid)

        rUrl = request.requestUrl()
        path = rUrl.path()
        ipfsPathS = joinIpfs(self.proxied.dagCid) + path

        ipfsPath = IPFSPath(ipfsPathS)
        if not ipfsPath.valid:
            return self.urlInvalid(request)

        return await self.fetchFromPath(ipfsop, request, ipfsPath, uid)


class GalacteekSchemeHandler(DAGProxySchemeHandler):
    @ipfsOp
    async def getDag(self, ipfsop):
        curProfile = ipfsop.ctx.currentProfile
        if curProfile:
            return curProfile.dagUser


class DAGWatchSchemeHandler(DAGProxySchemeHandler):
    def __init__(self, app, dappName, parent=None):
        super(DAGWatchSchemeHandler, self).__init__(app, parent=parent)

        self.name = dappName
        self.app.towers['dags'].dappDeployedAtCid.connect(self.onDappDeployed)

    def onDappDeployed(self, dag, name, cid):
        if name == self.name:
            self.proxied = dag

    @ipfsOp
    async def getDag(self, ipfsop):
        return self.proxied


class DWebSchemeHandler(NativeIPFSSchemeHandler):
    """
    IPFS dweb scheme handler, supporting URLs such as:

        dweb:/ipfs/multihash
        dweb:/ipns/domain.com/path
        dweb://ipfs/multihash/...
        fs:/ipfs/multihash
        etc..
    """

    @ipfsOp
    async def handleRequest(self, ipfsop, request, uid):
        rUrl = request.requestUrl()
        ipfsPath = IPFSPath(rUrl.toString())

        if not ipfsPath.valid:
            return self.urlInvalid(request)

        try:
            with async_timeout.timeout(self.requestTimeout):
                return await self.fetchFromPath(ipfsop, request, ipfsPath, uid)
        except asyncio.TimeoutError:
            return self.reqFailed(request)
        except Exception:
            # Any other error
            return self.reqFailed(request)


class MultiDAGProxySchemeHandler(NativeIPFSSchemeHandler):
    """
    Proxy handler to a stack of DAGs (first DAG registered is the
    first DAG tried).
    """

    def __init__(self, app, parent=None):
        super(MultiDAGProxySchemeHandler, self).__init__(app, parent=parent)
        self.proxied = []

    def useDag(self, dag):
        self.proxied.append(dag)

    async def renderDagListing(self, request, client, path, **kw):
        dag = kw.pop('dag', None)
        assert dag is not None

        listing = await dag.list(
            path=path.subPath if path.subPath else '')
        if not listing:
            return b'ELISTING'

        ctx = {}
        ctx['path'] = str(path)
        ctx['links'] = []

        for nname in listing:
            child = path.child(nname)
            ctx['links'].append({
                'name': nname,
                'href': child.ipfsUrl
            })

        try:
            data = await renderTemplate('ipfsdagnodelisting.html', **ctx)
        except:
            pass
        else:
            return data.encode()

    @ipfsOp
    async def handleRequest(self, ipfsop, request, uid):
        if len(self.proxied) == 0:
            return self.urlInvalid(request)

        rUrl = request.requestUrl()
        path = rUrl.path()

        for dag in self.proxied:
            if not dag.dagCid:
                continue

            ipfsPathS = joinIpfs(dag.dagCid) + path
            ipfsPath = IPFSPath(ipfsPathS)

            if not ipfsPath.valid:
                return request.fail(QWebEngineUrlRequestJob.UrlInvalid)

            return await self.fetchFromPath(ipfsop, request, ipfsPath, uid,
                                            dag=dag)

    async def fetchFromPath(self, ipfsop, request, ipfsPath, uid, **kw):
        dag = kw.pop('dag', None)
        assert dag is not None

        log.debug('Multi DAG proxy : Fetch-from-path {}'.format(ipfsPath))
        try:
            data = await ipfsop.client.cat(ipfsPath.objPath)
        except aioipfs.APIError as exc:
            await asyncio.sleep(0)

            self.debug('API error ({path}): {err}'.format(
                path=str(ipfsPath), err=exc.message)
            )

            dec = APIErrorDecoder(exc)

            if dec.errNoSuchLink():
                return self.urlNotFound(request)

            if dec.errIsDirectory():
                data = await self.renderDirectory(
                    request,
                    ipfsop.client,
                    str(ipfsPath)
                )
                if data:
                    return await self.renderData(request, ipfsPath, data, uid)

            if dec.errUnknownNode():
                # UNK DAG / TODO
                data = await self.renderDagNode(
                    request,
                    ipfsop.client,
                    ipfsPath,
                    dag=dag
                )
                if data:
                    return await self.renderData(request, ipfsPath, data, uid)
        else:
            return await self.renderData(request, ipfsPath, data, uid)


class EthDNSProxySchemeHandler(NativeIPFSSchemeHandler, IPFSObjectProxyScheme):
    """
    ENS scheme handler (resolves ENS domains through EthDNS with aiodns)

    This does not redirect to the IPFS object, but rather acts as a proxy.
    """

    def __init__(self, app, resolver=None, parent=None, cacheMaxEntries=64):
        super(EthDNSProxySchemeHandler, self).__init__(app, parent)

        self.ethResolver = resolver if resolver else \
            EthDNSResolver(self.app.loop)
        self._ensCache = collections.OrderedDict()
        self._ensCacheMaxEntries = cacheMaxEntries
        self._ensExpires = 60 * 10

    def debug(self, msg):
        log.debug('EthDNS proxy: {}'.format(msg))

    async def urlProxiedPath(self, rUrl):
        domain = rUrl.host()
        uPath = rUrl.path()

        if not domain or len(domain) > 512 or not domainValid(domain):
            return None

        now = time.time()
        lastDnsLink = lastResolved = None
        cached = self._ensCache.get(domain, None)

        if cached:
            lastDnsLink = cached['dnslinkp']
            lastResolved = cached['rlast']

        linkExpired = lastResolved and (now - lastResolved) > self._ensExpires

        if not lastDnsLink or linkExpired:
            logUser.info('EthDNS: resolving {0}'.format(domain))
            path = await self.ethResolver.resolveEnsDomain(domain)
        else:
            path = lastDnsLink

        if path and path.valid:
            if not cached:
                if len(self._ensCache) >= self._ensCacheMaxEntries:
                    self._ensCache.popitem(last=False)

                self._ensCache[domain] = {
                    'dnslinkp': path,
                    'rlast': now
                }

            return path.child(uPath) if uPath else path

    @ipfsOp
    async def handleRequest(self, ipfsop, request, uid):
        rUrl = request.requestUrl()
        domain = rUrl.host()

        if not domain or len(domain) > 512 or not domainValid(domain):
            logUser.info('EthDNS: invalid domain request')
            return self.urlInvalid(request)

        path = await self.urlProxiedPath(rUrl)

        if path and path.valid:
            return await self.fetchFromPath(ipfsop, request, path, uid)
        else:
            logUser.info('EthDNS: {domain} resolve failed'.format(
                domain=domain))
            return self.urlNotFound(request)



import os.path
import uuid
import pkg_resources
from datetime import datetime
from datetime import timezone

from feedgen.feed import FeedGenerator

from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal

from galacteek import log
from galacteek import logUser
from galacteek import ensure
from galacteek.ipfs import ipfsOp
from galacteek.ipfs.stat import StatInfo
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.cidhelpers import joinIpns
from galacteek.core.analyzer import ResourceAnalyzer
from galacteek.core import isoformat
from galacteek.dweb import render
from galacteek.dweb.atom import DWEB_ATOM_FEEDFN

from galacteek.ipfs.dag import EvolvingDAG


POST_NODEKEY = '_post'
CATEGORIES_NODEKEY = '_categories'
TAGS_NODEKEY = '_tags'
METADATA_NODEKEY = '_metadata'
BLOG_NODEKEY = 'blog'
PINREQS_NODEKEY = 'pinrequests'


class UserDAG(EvolvingDAG):
    def updateDagSchema(self, root):
        changed = False

        if BLOG_NODEKEY not in root:
            root[BLOG_NODEKEY] = {
                TAGS_NODEKEY: {}
            }
            changed = True

        if PINREQS_NODEKEY not in root:
            root[PINREQS_NODEKEY] = []
            changed = True

        if METADATA_NODEKEY not in root:
            root[METADATA_NODEKEY] = {
                'site_name': None,
                'site_description': None,
                'date_updated': None
            }
            changed = True

        return changed

    def initDag(self):
        return {
            'index.html': 'Blank',
            'media': {
                'images': {},
            }
        }

    async def siteMetadata(self):
        return await self.get(METADATA_NODEKEY)

    async def neverUpdated(self):
        metadata = await self.siteMetadata()
        return metadata['date_updated'] is None


class UserWebsite(QObject):
    """
    The website associated to the user.

    Later on this should become more generic to create all kinds of apps
    """

    websiteUpdated = pyqtSignal()

    def __init__(self, edag, profile, ipnsKeyId, jinjaEnv, parent=None):
        super().__init__(parent)

        self._profile = profile
        self._ipnsKeyId = ipnsKeyId
        self._edag = edag
        self._jinjaEnv = jinjaEnv

        self._atomFeedFn = DWEB_ATOM_FEEDFN

        self._rscAnalyzer = ResourceAnalyzer(None)
        self._updating = False

    @property
    def edag(self):
        return self._edag

    @property
    def profile(self):
        return self._profile

    @property
    def updating(self):
        return self._updating

    @property
    def ipnsKey(self):
        return self._ipnsKeyId

    @property
    def siteUrl(self):
        return self.sitePath.dwebUrl

    @property
    def sitePath(self):
        return IPFSPath(joinIpns(self.ipnsKey))

    @property
    def atomFeedPath(self):
        return self.sitePath.child(self._atomFeedFn)

    @property
    def rssFeedPath(self):
        return self.sitePath.child('rss.xml')

    @property
    def dagUser(self):
        return self.profile.dagUser

    @property
    def dagRoot(self):
        return self.profile.dagUser.root

    @property
    def dagRequests(self):
        return self.dagRoot['pinrequests']

    @property
    def dagBlog(self):
        return self.dagRoot['blog']

    def debug(self, msg):
        return self.profile.debug(msg)

    @ipfsOp
    async def blogEntries(self, op):
        return await self.edag.list(path='blog')

    @ipfsOp
    async def blogPost(self, ipfsop, title, msg, category=None, author=None):
        async with self.edag:
            username = self.profile.userInfo.username
            uid = str(uuid.uuid4())
            postName = title.strip().lower().replace(' ', '-')

            exEntries = await self.blogEntries()
            if postName in exEntries:
                postName += uid[0:8]
                return False

            now = datetime.now(timezone.utc)

            postObject = {
                'blogpost': {
                    'body': msg,
                    'title': title,
                    'uuid': uid,
                    'postname': postName,  # name of the post's DAG node
                    'tags': [],
                    'category': 'test',
                    'author': author if author else username,
                    'date_published': isoformat(now),
                    'date_modified': isoformat(now)
                }
            }

            # Create the DAG node for this post
            # Its view will be rendered later on

            self.dagBlog[postName] = {
                POST_NODEKEY: postObject
            }

        await ipfsop.sleep(2)
        await self.update()

        logUser.info('Your blog post is online')
        return True

    @ipfsOp
    async def makePinRequest(self, op, title, descr, objects,
                             priority=0, tags=None):
        async with self.profile.dagUser as dag:
            # objects needs to be a list of IPFS objects paths

            if not isinstance(objects, list):
                return False

            username = self.profile.userInfo.username
            uid = str(uuid.uuid4())
            now = datetime.now(timezone.utc)

            objsReq = []

            for obj in objects:
                path = IPFSPath(obj)
                if not path.valid:
                    continue

                try:
                    mType, stat = await self._rscAnalyzer(path)
                except:
                    continue

                statInfo = StatInfo(stat)
                if not statInfo.valid:
                    continue

                objsReq.append({
                    'path': path.objPath,
                    'totalsize': statInfo.totalSize,
                    'content_type': str(mType) if mType else None
                })

            pinRequest = {
                'pinrequest': {
                    'body': None,
                    'title': title,
                    'description': descr,
                    'objects': objsReq,
                    'uuid': uid,
                    'tags': tags if tags else [],
                    'author': username,
                    'priority': priority,
                    'date_published': isoformat(now),
                    'version': 1
                }
            }

            dag.dagRoot[PINREQS_NODEKEY].append(pinRequest)

        await self.update()
        return True

    @ipfsOp
    async def init(self, ipfsop):
        # Import the assets

        assetsPath = pkg_resources.resource_filename('galacteek.templates',
                                                     'usersite/assets')
        self.assetsEntry = await ipfsop.addPath(assetsPath, recursive=True)

        if self.profile.userInfo.avatarCid:
            self.edag.root['media']['images']['avatar'] = self.edag.mkLink(
                self.profile.userInfo.avatarCid)

        if self.profile.ctx.hasRsc('ipfs-cube-64'):
            self.edag.root['media']['images']['ipfs-cube.png'] = \
                self.edag.mkLink(self.profile.ctx.resources['ipfs-cube-64'])
        if self.profile.ctx.hasRsc('atom-feed'):
            self.edag.root['media']['images']['atom-feed.png'] = \
                self.edag.mkLink(self.profile.ctx.resources['atom-feed'])

        self.edag.root['about.html'] = await self.renderLink(
            'usersite/about.html')

    def createFeed(self):
        # Create the feed

        feed = FeedGenerator()
        feed.id(self.siteUrl)
        feed.title("{0}'s dweb space".format(
            self.profile.userInfo.username))
        feed.author({
            'name': self.profile.userInfo.username,
            'email': self.profile.userInfo.email
        })
        feed.link(href=self.atomFeedPath.dwebUrl, rel='self')
        feed.language('en')
        return feed

    @ipfsOp
    async def update(self, op):
        if self.dagRoot is None or self.updating is True:
            return

        now = datetime.now(timezone.utc)
        self._updating = True

        assetsChanged = False
        blogPosts = []

        feed = self.createFeed()

        async with self.profile.dagUser as dag:
            if self.assetsEntry:
                resolved = await dag.resolve('assets')
                if resolved and self.assetsEntry.get('Hash') != resolved:
                    # The assets were modified (will need to fully regen)
                    assetsChanged = True

                dag.root['assets'] = dag.mkLink(self.assetsEntry)

            entries = await dag.list(path='blog')
            bEntries = entries if isinstance(entries, list) else []

            for entry in bEntries:
                await op.sleep()

                if entry == TAGS_NODEKEY:
                    continue

                nList = await dag.list(path='blog/{}'.format(entry))

                postObj = await dag.get('blog/{0}/{1}'.format(
                    entry, POST_NODEKEY))

                if not isinstance(postObj, dict):
                    continue

                post = postObj['blogpost']

                blogPosts.append(post)

                if 'view' in nList and assetsChanged is False:
                    continue

                self.dagBlog[entry]['view'] = await self.renderLink(
                    'usersite/blog_post.html',
                    contained=True,
                    post=post,
                )

            blogPosts = sorted(blogPosts,
                               key=lambda post: post['date_published'],
                               reverse=True)

            self.feedAddPosts(blogPosts, feed)

            requests = await dag.get(PINREQS_NODEKEY)

            if requests:
                for idx, request in enumerate(requests):
                    path = '{0}/{1}'.format(PINREQS_NODEKEY, idx)
                    if not await dag.get(
                            '{path}/view'.format(path=path, idx=idx)):
                        self.dagRequests[idx]['view'] = \
                            await self.renderLink(
                                'usersite/pinrequest.html',
                                request=request
                        )

                self.feedAddPinRequests(requests, feed)

            dag.root['pinrequests.html'] = await self.renderLink(
                'usersite/pinrequests.html',
                pinrequests=requests if requests else []
            )

            dag.root['index.html'] = await self.renderLink(
                'usersite/home.html',
                posts=blogPosts
            )

            dag.root['about.html'] = await self.renderLink(
                'usersite/about.html')

            # Finally, generate the Atom feed and put it in the graph
            try:
                atomFeed = feed.atom_str(pretty=True)
                entry = await op.addBytes(atomFeed)
                if entry:
                    # Unpin the old feed (disabled for now)
                    if self._atomFeedFn in dag.root and 0:
                        resolved = await dag.resolve(
                            path=self._atomFeedFn)
                        if resolved:
                            ensure(op.unpin(resolved))

                    dag.root[self._atomFeedFn] = dag.mkLink(entry)
            except Exception as err:
                log.debug('Error generating Atom feed: {}'.format(
                    str(err)))

            metadata = await dag.siteMetadata()
            metadata['date_updated'] = now.isoformat()
            dag.root[METADATA_NODEKEY] = metadata

        self.websiteUpdated.emit()
        self._updating = False

    def feedAddPinRequests(self, requests, feed):
        for id, reqobj in enumerate(requests):
            req = reqobj['pinrequest']
            rpath = self.sitePath.child(os.path.join(
                PINREQS_NODEKEY, str(id), 'view'))

            fEntry = feed.add_entry()
            fEntry.title('Pin request: {}'.format(req['title']))
            fEntry.id(rpath.dwebUrl)
            fEntry.link(href=rpath.dwebUrl, rel='alternate')
            fEntry.published(req['date_published'])
            fEntry.author({'name': self.profile.userInfo.username})

    def feedAddPosts(self, blogPosts, feed):
        for post in blogPosts:
            ppath = self.sitePath.child(os.path.join(
                'blog', post['postname'], 'view'))

            fEntry = feed.add_entry()
            fEntry.title(post['title'])
            fEntry.id(ppath.dwebUrl)
            fEntry.link(href=ppath.dwebUrl, rel='alternate')
            fEntry.updated(post['date_modified'])
            fEntry.published(post['date_published'])
            fEntry.author({'name': self.profile.userInfo.username})

            for tag in post['tags']:
                fEntry.category(term=tag)

    async def tmplRender(self, tmpl, contained=False, **kw):
        coro = render.ipfsRender if not contained else \
            render.ipfsRenderContained

        return await coro(self._jinjaEnv,
                          tmpl,
                          profile=self.profile,
                          dag=self.edag.dagRoot,
                          siteIpns=self.ipnsKey,
                          atomFeedrUrl=self.atomFeedPath.dwebUrl,
                          **kw)

    async def renderLink(self, tmpl, contained=False, **kw):
        # Render and link in the dag
        result = await self.tmplRender(tmpl, contained=contained, **kw)
        return self.edag.mkLink(result)



import os.path

from PyQt5.QtWidgets import QApplication

from PyQt5.QtGui import QStandardItemModel

from PyQt5.QtCore import QUrl
from PyQt5.QtCore import Qt
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QMimeData
from PyQt5.QtCore import QDir
from PyQt5.QtCore import QFile
from PyQt5.QtCore import QModelIndex

from galacteek.ipfs.cidhelpers import joinIpfs
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.appsettings import *

from galacteek.core.modelhelpers import *
from galacteek.ui.i18n import *  # noqa
from galacteek.ui.helpers import *


def sampleQrCodes():
    dir = QDir(':/share/qr-codes')
    if dir.exists():
        for entry in dir.entryList():
            yield entry, QFile(':/share/qr-codes/{}'.format(entry))


class MFSItem(UneditableItem):
    def __init__(self, text, path=None, parenthash=None, icon=None):
        super(MFSItem, self).__init__(text, icon=icon)
        self._path = path
        self._parentHash = parenthash

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, v):
        self._path = v

    @property
    def parentHash(self):
        return self._parentHash

    def setParentHash(self, pHash):
        self._parentHash = pHash

    def childrenItems(self):
        for row in range(0, self.rowCount()):
            child = self.child(row, 0)
            if child:
                yield child

    def findChildByMultihash(self, multihash):
        for item in self.childrenItems():
            if not isinstance(item, MFSNameItem):
                continue
            if item.entry['Hash'] == multihash:
                return item

    def findChildByName(self, name):
        for item in self.childrenItems():
            if not isinstance(item, MFSNameItem):
                continue
            if item.entry['Name'] == name:
                return item


class MFSRootItem(MFSItem):
    """
    Root item (top-level items in the MFS model)
    """

    expandedItemsCount = 0

    def __init__(self, text, path=None, parenthash=None, icon=None,
                 offline=False, alwaysOffline=False):
        super(MFSRootItem, self).__init__(text, path=path,
                                          parenthash=parenthash,
                                          icon=icon)

        # alwaysOffline won't be used by the filemanager for now
        self.offline = True if alwaysOffline else offline
        self.alwaysOffline = alwaysOffline


class MFSNameItem(MFSItem):
    def __init__(self, entry, text, icon):
        super().__init__(text, icon=icon)

        self._entry = entry
        self._mimeType = None

    @property
    def entry(self):
        return self._entry

    @property
    def ipfsPath(self):
        return IPFSPath(self.fullPath)

    @property
    def dwebUrl(self):
        return QUrl('dweb:{}'.format(self.fullPath))

    @property
    def mimeType(self):
        return self._mimeType

    def mimeFromDb(self, db):
        self._mimeType = db.mimeTypeForFile(self.entry['Name'])

    @property
    def mimeCategory(self):
        if self.mimeType:
            return self.mimeTypeName.split('/')[0]

    @property
    def mimeTypeName(self):
        if self.mimeType:
            return self.mimeType.name()

    @property
    def fullPath(self):
        parentHash = self.parentHash
        if parentHash:
            fp = joinIpfs(os.path.join(parentHash, self.entry['Name']))
        else:
            fp = joinIpfs(self.entry['Hash'])

        return fp + '/' if self.isDir() else fp

    def isFile(self):
        return self.entry['Type'] == 0

    def isDir(self):
        return self.entry['Type'] == 1


class MFSItemModel(QStandardItemModel):
    # Drag and drop events
    fileDropEvent = pyqtSignal(str)
    directoryDropEvent = pyqtSignal(str)

    # An item's view was refreshed
    refreshed = pyqtSignal(MFSNameItem)

    needExpand = pyqtSignal(QModelIndex)

    def __init__(self):
        QStandardItemModel.__init__(self)
        self.app = QApplication.instance()

        self.itemRoot = self.invisibleRootItem()
        self.itemRootIdx = self.indexFromItem(self.itemRoot)
        self.initialized = False
        self.qrInitialized = False

    def setupItemsFromProfile(self, profile):
        self.itemHome = MFSRootItem(iHome(),
                                    path=profile.pathHome,
                                    icon=getIcon('go-home.png'))
        self.itemImages = MFSRootItem(iImages(),
                                      path=profile.pathImages,
                                      icon=getMimeIcon('image/x-generic'))
        self.itemPictures = MFSRootItem(iPictures(),
                                        path=profile.pathPictures,
                                        icon=getIcon('folder-pictures.png'))
        self.itemVideos = MFSRootItem(iVideos(),
                                      path=profile.pathVideos,
                                      icon=getIcon('folder-videos.png'))
        self.itemMusic = MFSRootItem(iMusic(),
                                     path=profile.pathMusic,
                                     icon=getIcon('folder-music.png'))
        self.itemCode = MFSRootItem(iCode(),
                                    path=profile.pathCode,
                                    icon=getIcon('code-fork.png'))
        self.itemDocuments = MFSRootItem(iDocuments(),
                                         path=profile.pathDocuments,
                                         icon=getIcon('folder-documents.png'))
        self.itemWebPages = MFSRootItem(iWebPages(),
                                        path=profile.pathWebPages,
                                        icon=getMimeIcon('text/html'))
        self.itemDWebApps = MFSRootItem(iDWebApps(),
                                        path=profile.pathDWebApps,
                                        icon=getIcon('distributed.png'))
        self.itemQrCodes = MFSRootItem(iQrCodes(),
                                       path=profile.pathQrCodes,
                                       icon=getIcon('ipfs-qrcode.png'))
        self.itemTemporary = MFSRootItem(iTemporaryFiles(),
                                         alwaysOffline=True,
                                         path=profile.pathTmp,
                                         icon=getIcon('folder-temp.png'))
        self.itemEncrypted = MFSRootItem(iEncryptedFiles(),
                                         alwaysOffline=True,
                                         path=profile.pathEncryptedFiles,
                                         icon=getIcon('key-diago.png'))

        # Core MFS folders (visible in the "Link to MFS" folders)
        self.fsCore = [
            self.itemHome,
            self.itemTemporary,
            self.itemPictures,
            self.itemImages,
            self.itemVideos,
            self.itemCode,
            self.itemMusic,
            self.itemDocuments,
            self.itemWebPages,
            self.itemDWebApps
        ]

        # Extra folders
        self.fsExtra = [
            self.itemQrCodes,
            self.itemEncrypted
        ]

        self.filesystem = self.fsCore + self.fsExtra

        self.itemRoot.appendRows(self.filesystem)

        self.initialized = True

    @ipfsOp
    async def setupQrCodesFolder(self, ipfsop, profile, model, files):
        """
        Import the sample QR codes to the corresponding folder
        Needs rewrite ..
        """
        for name, qrFile in sampleQrCodes():
            await ipfsop.sleep()

            if qrFile.exists():
                path = self.app.tempDir.filePath(name)
                if qrFile.copy(path):
                    tmpFile = QFile(path)
                    entry = await ipfsop.addPath(path)
                    await ipfsop.sleep()

                    if not entry:
                        tmpFile.remove()
                        continue

                    exists = await ipfsop.filesLookupHash(
                        model.itemQrCodes.path, entry['Hash'])

                    await ipfsop.sleep()

                    if not exists:
                        await files.linkEntry(ipfsop, entry,
                                              model.itemQrCodes.path,
                                              entry['Name'])

                    tmpFile.remove()

        self.qrInitialized = True

    def displayItem(self, arg):
        self.itemRoot.appendRow(arg)

    def supportedDropActions(self):
        return Qt.CopyAction | Qt.MoveAction | \
            Qt.TargetMoveAction | Qt.LinkAction

    def mimeData(self, indexes):
        mimedata = QMimeData()

        for idx in indexes:
            if not idx.isValid():
                continue

            nameItem = self.getNameItemFromIdx(idx)

            if nameItem:
                mimedata.setUrls([nameItem.dwebUrl])
                break

        return mimedata

    def canDropMimeData(self, data, action, row, column, parent):
        mimeText = data.text()

        if mimeText and mimeText.startswith('file://'):
            return True
        return False

    def dropMimeData(self, data, action, row, column, parent):
        if data.hasUrls():
            for url in data.urls():
                path = url.toLocalFile()
                if os.path.isfile(path):
                    self.fileDropEvent.emit(path)
                if os.path.isdir(path):
                    self.directoryDropEvent.emit(path)
            return True
        else:
            return False

    def getHashFromIdx(self, idx):
        idxHash = self.index(idx.row(), 2, idx.parent())
        return self.data(idxHash)

    def getNameFromIdx(self, idx):
        idxName = self.index(idx.row(), 0, idx.parent())
        return self.data(idxName)

    def getNameItemFromIdx(self, idx):
        idxName = self.index(idx.row(), 0, idx.parent())
        return self.itemFromIndex(idxName)


def createMFSModel():
    # Setup the model
    model = MFSItemModel()
    model.setHorizontalHeaderLabels(
        [iFileName(), iFileSize(), iCID()])
    return model



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/batchpinlist.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_BatchPinList(object):
    def setupUi(self, BatchPinList):
        BatchPinList.setObjectName("BatchPinList")
        BatchPinList.resize(588, 397)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(BatchPinList)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.labelBasePath = QtWidgets.QLabel(BatchPinList)
        self.labelBasePath.setText("")
        self.labelBasePath.setObjectName("labelBasePath")
        self.horizontalLayout.addWidget(self.labelBasePath)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.tableWidget = QtWidgets.QTableWidget(BatchPinList)
        self.tableWidget.setShowGrid(False)
        self.tableWidget.setColumnCount(4)
        self.tableWidget.setObjectName("tableWidget")
        self.tableWidget.setRowCount(0)
        self.tableWidget.horizontalHeader().setVisible(True)
        self.tableWidget.horizontalHeader().setStretchLastSection(True)
        self.verticalLayout.addWidget(self.tableWidget)
        self.hLayoutCtrl = QtWidgets.QHBoxLayout()
        self.hLayoutCtrl.setObjectName("hLayoutCtrl")
        self.proceedButton = QtWidgets.QPushButton(BatchPinList)
        self.proceedButton.setMaximumSize(QtCore.QSize(300, 16777215))
        self.proceedButton.setObjectName("proceedButton")
        self.hLayoutCtrl.addWidget(self.proceedButton)
        self.cancelButton = QtWidgets.QPushButton(BatchPinList)
        self.cancelButton.setMaximumSize(QtCore.QSize(200, 16777215))
        self.cancelButton.setObjectName("cancelButton")
        self.hLayoutCtrl.addWidget(self.cancelButton)
        self.verticalLayout.addLayout(self.hLayoutCtrl)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(BatchPinList)
        QtCore.QMetaObject.connectSlotsByName(BatchPinList)

    def retranslateUi(self, BatchPinList):
        _translate = QtCore.QCoreApplication.translate
        BatchPinList.setWindowTitle(_translate("BatchPinList", "Form"))
        self.proceedButton.setText(_translate("BatchPinList", "Pin selection"))
        self.cancelButton.setText(_translate("BatchPinList", "Cancel"))




from PyQt5.QtCore import QCoreApplication

from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.cidhelpers import *

from .helpers import *
from .dialogs import *
from .widgets import *
from .i18n import *


def iPath():
    return QCoreApplication.translate('HashmarksViewForm', 'Path')


def iTitle():
    return QCoreApplication.translate('HashmarksViewForm', 'Title')


def iShared():
    return QCoreApplication.translate('HashmarksViewForm', 'Shared')


def iDate():
    return QCoreApplication.translate('HashmarksViewForm', 'Date')


def iTimestamp():
    return QCoreApplication.translate('HashmarksViewForm', 'Timestamp')


def iAlreadyHashmarked():
    return QCoreApplication.translate('HashmarksViewForm',
                                      'Already Hashmarked')


def iInvalidHashmarkPath():
    return QCoreApplication.translate('HashmarksViewForm',
                                      'Invalid hashmark path')


def iImportHashmark():
    return QCoreApplication.translate('HashmarksViewForm',
                                      'Import hashmark to')


def iNetworkMarks():
    return QCoreApplication.translate('HashmarksViewForm', 'Network hashmarks')


def iFeeds():
    return QCoreApplication.translate('HashmarksViewForm', 'Feeds')


def addHashmark(hashmarks, path, title, description='', stats={},
                pin=False, pinRecursive=False):
    if hashmarks.find(path):
        messageBox(iAlreadyHashmarked())
        return False

    ipfsPath = IPFSPath(path, autoCidConv=True)
    if not ipfsPath.valid:
        messageBox(iInvalidHashmarkPath())
        return False

    runDialog(AddHashmarkDialog, hashmarks,
              str(ipfsPath), title, description, stats,
              pin=pin, pinRecursive=pinRecursive)
    return True



from aiohttp import Signal

from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal  # noqa


class ASig(Signal):
    pass


class SingleASig(ASig):
    def __init__(self, callback, owner=None):
        super().__init__(owner)

        self.append(callback)
        self.freeze()


class GObject(QObject):
    pass



from PyQt5.QtWebEngineWidgets import QWebEngineScript
from PyQt5.QtCore import QFile


def scriptFromString(name, jsCode):
    script = QWebEngineScript()
    script.setName(name)
    script.setSourceCode(jsCode)
    script.setWorldId(QWebEngineScript.MainWorld)
    script.setInjectionPoint(QWebEngineScript.DocumentCreation)
    return script


ipfsInjScript = '''
window.ipfs = window.IpfsHttpClient('{host}', '{port}');
'''


def ipfsClientScripts(connParams):
    scripts = []
    jsFile = QFile(':/share/js/ipfs-http-client/index.min.js')
    if not jsFile.open(QFile.ReadOnly):
        return

    scriptJsIpfs = QWebEngineScript()
    scriptJsIpfs.setName('ipfs-http-client')

    libCode = jsFile.readAll().data().decode('utf-8')
    libCode += "\n"
    libCode += ipfsInjScript.format(
        host=connParams.host,
        port=connParams.apiPort)

    scriptJsIpfs.setSourceCode(libCode)
    scriptJsIpfs.setWorldId(QWebEngineScript.MainWorld)
    scriptJsIpfs.setInjectionPoint(QWebEngineScript.DocumentCreation)
    scriptJsIpfs.setRunsOnSubFrames(True)
    scripts.append(scriptJsIpfs)
    return scripts


w3ScriptHttp = '''
window.web3 = new Web3(
    new Web3.providers.HttpProvider('{rpcurl}')
);
window.web3.eth.defaultAccount = window.web3.eth.accounts[0];
'''

w3ScriptWs = '''
window.web3 = new Web3(
    new Web3.providers.WebsocketProvider('{rpcurl}')
);
window.web3.eth.defaultAccount = window.web3.eth.accounts[0];
'''


def ethereumClientScripts(connParams):
    scripts = []
    jsFile = QFile(':/share/js/web3.min.js')
    if not jsFile.open(QFile.ReadOnly):
        return None

    libCode = jsFile.readAll().data().decode('utf-8')
    libCode += "\n"

    if connParams.provType == 'http':
        libCode += w3ScriptHttp.format(rpcurl=connParams.rpcUrl)
    elif connParams.provType == 'websocket':
        libCode += w3ScriptWs.format(rpcurl=connParams.rpcUrl)

    scriptWeb3 = QWebEngineScript()
    scriptWeb3.setName('web3.js')
    scriptWeb3.setSourceCode(libCode)
    scriptWeb3.setWorldId(QWebEngineScript.MainWorld)
    scriptWeb3.setInjectionPoint(QWebEngineScript.DocumentCreation)
    scripts.append(scriptWeb3)
    return scripts


def orbitScripts(connParams):
    scripts = []
    jsFile = QFile(':/share/js/orbit-db/orbitdb.js')
    if not jsFile.open(QFile.ReadOnly):
        return
    scriptJsIpfs = QWebEngineScript()
    scriptJsIpfs.setName('orbit-db')
    scriptJsIpfs.setSourceCode(jsFile.readAll().data().decode('utf-8'))
    scriptJsIpfs.setWorldId(QWebEngineScript.MainWorld)
    scriptJsIpfs.setInjectionPoint(QWebEngineScript.DocumentCreation)
    scriptJsIpfs.setRunsOnSubFrames(True)
    scripts.append(scriptJsIpfs)

    script = QWebEngineScript()
    script.setSourceCode('\n'.join([
        "document.addEventListener('DOMContentLoaded', function () {",
        "window.orbitdb = new OrbitDB(window.ipfs)",
        "})"]))
    script.setWorldId(QWebEngineScript.MainWorld)
    return scripts



import re
import os.path
import os
import aioipfs

from PyQt5.QtCore import QUrl

from galacteek import log
from galacteek.ipfs.cid import make_cid
from galacteek.ipfs.cid import BaseCID
from galacteek.ipfs.stat import StatInfo

import multihash


def normp(path):
    sep = '/'
    final = path.endswith(sep)
    normed = os.path.normpath(path)

    if final:
        return normed + sep

    return normed


def normpPreserve(path, preserveTrailing=True):
    """
    Like os.path.normpath but preserves trailing slash by default
    and couldn't care less about normalizing ..
    """
    path = os.fspath(path)

    sep = '/'
    empty = ''
    dot = '.'

    if path == empty:
        return dot

    initial_slashes = path.startswith(sep)
    final_slashes = path.endswith(sep)

    comps = path.split(sep)
    new_comps = []
    for comp in comps:
        if comp in (empty, dot):
            continue
        new_comps.append(comp)
    comps = new_comps
    path = sep.join(comps)
    if initial_slashes:
        path = sep + path
    if final_slashes and preserveTrailing:
        path = path + sep
    return path or dot


def joinIpfs(path):
    if isinstance(path, str):
        return os.path.join('/ipfs/', path)


def joinIpns(path):
    if isinstance(path, str):
        return os.path.join('/ipns/', path)


def stripIpfs(path):
    if isinstance(path, str):
        return re.sub('^/ipfs/', '', path)


def stripIpns(path):
    if isinstance(path, str):
        return re.sub('^/ipns/', '', path)


def isIpfsPath(path):
    if isinstance(path, str):
        return ipfsRegSearchPath(path) is not None


def isIpnsPath(path):
    if isinstance(path, str):
        return ipnsRegSearchPath(path) is not None


def shortCidRepr(cid):
    cidStr = str(cid)
    if cid.version == 0:
        return '... {0}'.format(cidStr[3 * int(len(cidStr) / 5):])
    else:
        b32Cid = cidConvertBase32(cidStr)
        return '... {0}'.format(b32Cid[4 * int(len(b32Cid) / 5):])


def shortPathRepr(path):
    if isIpfsPath(path) or isIpnsPath(path):
        basename = os.path.basename(path)
        if cidValid(basename):
            cid = getCID(basename)
            return shortCidRepr(cid)
        else:
            return basename
    else:
        cid = getCID(path)
        if cid:
            return shortCidRepr(cid)
        else:
            return path


def isMultihash(hashstring):
    """
    Check if hashstring is a valid base58-encoded multihash

    :param str hashstring: the multihash to validate
    :return: if the value is a valid multihash or not
    :rtype: bool
    """
    try:
        multihash.decode(hashstring.encode('ascii'), 'base58')
        return True
    except BaseException:
        return False


def getCID(cidStr):
    try:
        return make_cid(cidStr)
    except BaseException:
        return None


def cidUpgrade(cid):
    """
    Converts a cid.CIDv0 instance to a cid.CIDv1
    """
    if issubclass(cid.__class__, BaseCID) and cid.version == 0:
        return cid.to_v1()


def cidConvertBase32(cid):
    """
    Convert a base58-encoded CID to a base32 string

    If it's a CIDv0, it's upgraded to CIDv1 first

    :param str cid: The cid string
    :rtype: str
    """

    cid = getCID(cid)
    if not cid:
        return None

    if cid.version == 0:
        if not cidValid(cid):
            return None

        cid = cidUpgrade(cid)
        if not cid:
            log.debug('Impossible to convert CIDv0: {}'.format(cid))
            return None

    return cid.encode('base32').decode()


def cidValid(cid):
    """
    Check if the passed argument is a valid IPFS CID

    :param cid: the CID to validate, can be a string or a BaseCID
    :return: if the value is a valid CID or not
    :rtype: bool
    """

    if cid is None:
        return False

    if isinstance(cid, str):
        c = getCID(cid)
    elif issubclass(cid.__class__, BaseCID):
        c = cid
    else:
        return False

    if c is None:
        return False
    if c.version in [0, 1]:
        # Ensure that we can decode the multihash
        try:  # can raise ValueError
            if multihash.decode(c.multihash):
                return True
        except BaseException:
            return False
    return False


def domainValid(domain):
    domainP = r'^([A-Za-z0-9]\.|[A-Za-z0-9][A-Za-z0-9-]{0,61}[A-Za-z0-9]\.){1,3}[A-Za-z]{2,6}$'  # noqa
    return re.match(domainP, domain)


# Regexps

ipfsPathRe = re.compile(
    r'^(\s*)?(?:fs:|dweb:|https?://[\w:.-]+)?(?P<fullpath>/ipfs/(?P<rootcid>[a-zA-Z0-9]{46,59})/?(?P<subpath>[\w<>":;,?!\*%&=@\$~/\s\.\-_\\\'()\+]{1,1024})?)#?(?P<fragment>[\w_\.\-\+,=/]{1,256})?$',  # noqa
    flags=re.UNICODE)

# For ipfs://<cid-base32>
ipfsPathDedRe = re.compile(
    r'^(\s*)?(?:ipfs://)(?P<fullpath>(?P<rootcid>[a-z2-7]{59})/?(?P<subpath>[\w<>"*:;,?!%&=@\$~/\s\.\-_\\\'\+()]{1,1024})?)#?(?P<fragment>[\w_\+\.\-,=/]{1,256})?$',  # noqa
    flags=re.UNICODE)

# For rewriting (unlawful) ipfs://<cidv0> or ipfs://<cidv1-base58> to base32
ipfsPathDedRe58 = re.compile(
    r'^(\s*)?(?:ipfs://)(?P<fullpath>(?P<rootcid>[a-zA-Z0-9]{46,59})/?(?P<subpath>[\w<>"*:;,?!%&=@\$~/\s\.\-_\'\\\+()]{1,1024})?)#?(?P<fragment>[\w\-\+_\.,=/]{1,256})?$',  # noqa
    flags=re.UNICODE)

ipnsPathDedRe = re.compile(
    r'^(\s*)?(?:(ipns|ipfs)://)(?P<fullpath>(?P<fqdn>[\w.-]+)/?(?P<subpath>[\w<>"*:;,\\?!%&=@\$~/\s\.\-_\'\+()]{1,1024})?)#?(?P<fragment>[\w\-\+_\.,=/]{1,256})?$',  # noqa
    flags=re.UNICODE)

ipfsCidRe = re.compile(
    r'^(\s*)?(?P<cid>[a-zA-Z0-9]{46,59})$')

ipfsCid32Re = re.compile(
    r'^(\s*)?(?P<cid>[a-z2-7]{59})$')

ipnsPathRe = re.compile(
    r'^(\s*)?(?:fs:|dweb:|https?://[\w:.-]+)?(?P<fullpath>/ipns/(?P<fqdn>[\w\.-]+)/?(?P<subpath>[\w<>"*:;,?!%&=@\$~/\s\.\-_\'\\\+()]{1,1024})?)#?(?P<fragment>[\w\+\-_\.,=/]{1,256})?$',  # noqa
    flags=re.UNICODE)


class IPFSPath:
    maxLength = 1024

    def __init__(self, input, autoCidConv=False, enableBase32=True):
        self._enableBase32 = enableBase32
        self._autoCidConv = autoCidConv
        self._rootCid = None
        self._rootCidV = None
        self._rootCidUseB32 = False
        self._input = input
        self._rscPath = None
        self._subPath = None
        self._fragment = None
        self._scheme = None
        self._resolvedCid = None
        self._ipnsId = None
        self._valid = self.__analyze()

    @property
    def autoCidConv(self):
        return self._autoCidConv

    @property
    def input(self):
        return self._input

    @property
    def valid(self):
        return self._valid

    @property
    def scheme(self):
        return self._scheme

    @property
    def ipnsId(self):
        return self._ipnsId

    @property
    def ipnsFqdn(self):
        if domainValid(self.ipnsId):
            return self.ipnsId

    @property
    def ipnsKey(self):
        if not self.ipnsFqdn:
            return self.ipnsId

    @property
    def resolvedCid(self):
        return self._resolvedCid

    @property
    def rootCidV(self):
        return self._rootCidV

    @property
    def rootCid(self):
        return self._rootCid

    @property
    def rootCidRepr(self):
        if self.rootCidUseB32:
            return self._rootCid.encode('base32').decode()
        else:
            return str(self._rootCid)

    @property
    def rootCidUseB32(self):
        return self._rootCidUseB32

    @property
    def isIpfs(self):
        return self.valid is True and self.scheme == 'ipfs'

    @property
    def isRoot(self):
        return self.isIpfsRoot or self.isIpnsRoot

    @property
    def isIpfsRoot(self):
        return self.isIpfs is True and self.subPath is None

    @property
    def isIpns(self):
        return self.valid is True and self.scheme == 'ipns'

    @property
    def isIpnsRoot(self):
        return self.isIpns and self.subPath is None

    @property
    def fragment(self):
        # Return the fragment if there's any
        return self._fragment

    @fragment.setter
    def fragment(self, frag):
        self._fragment = frag

    @property
    def path(self):
        # Return the object path (without fragment)
        return self.objPath

    @property
    def basename(self):
        if self.valid:
            return os.path.basename(self.objPath.rstrip('/'))

    @property
    def objPath(self):
        # Return the object path (without fragment)
        return self._rscPath

    @property
    def subPath(self):
        # Return the sub path
        return self._subPath

    @property
    def fullPath(self):
        if isinstance(self.fragment, str):
            return self._rscPath + '#' + self.fragment
        else:
            return self._rscPath

    @property
    def dwebUrl(self):
        return '{scheme}:{path}'.format(
            scheme='dweb',
            path=self.fullPath
        )

    @property
    def ipfsUrl(self):
        if self.isIpfs and self.rootCidUseB32:
            # ipfs://
            return '{scheme}://{path}'.format(
                scheme=self.scheme,
                path=stripIpfs(self.fullPath)
            )
        elif self.isIpns:
            if domainValid(self._ipnsId):
                return '{scheme}://{path}'.format(
                    scheme='ipns',
                    path=stripIpns(self.fullPath)
                )
            else:
                return self.dwebUrl
        else:
            return self.dwebUrl

    @property
    def dwebQtUrl(self):
        return QUrl(self.dwebUrl)

    def __analyze(self):
        """
        Analyze the path and returns a boolean (valid or not)

        :rtype bool
        """

        if not isinstance(self.input, str):
            return False

        self._input = self._input.strip()

        if len(self.input) > self.maxLength:
            return False

        ma = ipfsDedSearchPath(self.input)
        if ma:
            gdict = ma.groupdict()
            if 'rootcid' not in gdict or 'fullpath' not in gdict:
                return False

            cid = ma.group('rootcid')
            if not cidValid(cid):
                return False

            if not self.parseCid(cid):
                return False

            subpath = gdict.get('subpath')

            if subpath:
                self._rscPath = os.path.join(
                    joinIpfs(self.rootCidRepr),
                    subpath
                )
            else:
                self._rscPath = joinIpfs(self.rootCidRepr)

            self._fragment = gdict.get('fragment')
            self._subPath = subpath
            self._scheme = 'ipfs'
            return True

        ma = ipfsRegSearchPath(self.input)
        if ma:
            gdict = ma.groupdict()
            if 'rootcid' not in gdict or 'fullpath' not in gdict:
                return False

            cid = ma.group('rootcid')
            if not cidValid(cid):
                return False

            if not self.parseCid(cid):
                return False

            subpath = gdict.get('subpath')
            if subpath:
                self._rscPath = os.path.join(
                    joinIpfs(self.rootCidRepr),
                    subpath
                )
            else:
                self._rscPath = joinIpfs(self.rootCidRepr)

            self._fragment = gdict.get('fragment')
            self._subPath = gdict.get('subpath')
            self._scheme = 'ipfs'
            return True

        ma = ipnsRegSearchPath(self.input)
        if ma:
            gdict = ma.groupdict()

            subpath = gdict.get('subpath')
            if subpath:
                self._rscPath = os.path.join(
                    joinIpns(gdict.get('fqdn')),
                    subpath
                )
            else:
                self._rscPath = joinIpns(gdict.get('fqdn'))

            self._ipnsId = gdict.get('fqdn')
            self._fragment = gdict.get('fragment')
            self._subPath = gdict.get('subpath')
            self._scheme = 'ipns'
            return True

        ma = ipfsRegSearchCid(self.input)
        if ma:
            cidStr = ma.group('cid')
            if not cidValid(cidStr):
                return False

            if not self.parseCid(cidStr):
                return False

            self._rscPath = joinIpfs(self.rootCidRepr)
            self._scheme = 'ipfs'
            return True

        return False

    def parseCid(self, cidStr):
        if not cidValid(cidStr):
            return False

        self._rootCid = getCID(cidStr)
        if self.rootCid:
            self._rootCidV = self.rootCid.version if \
                self.rootCid.version in range(0, 2) else None
        else:
            return False

        if self._autoCidConv and self.rootCid.version == 0:
            # Automatic V0-to-V1 conversion
            self._rootCid = cidUpgrade(self._rootCid)

        if self.rootCid.version == 1 and self._enableBase32:
            # rootCidRepr will convert it to base32
            self._rootCidUseB32 = True

        return True

    def child(self, path, normalize=False):
        if not isinstance(path, str):
            raise ValueError('Need string')

        if self.subPath is None and (path == '..' or path.startswith('../')):
            # Not crossing the /{ipfs,ipns} NS
            return self

        if normalize:
            cPath = normp(path)
            return IPFSPath(normp(
                os.path.join(self.objPath, cPath.lstrip('/'))))
        else:
            cPath = normpPreserve(path)
            return IPFSPath(os.path.join(self.objPath, cPath.lstrip('/')))

    def parent(self):
        return self.child('../', normalize=True)

    def root(self):
        n = self
        while n.valid and n.subPath:
            n = n.parent()

        return n if n.valid else None

    def shortRepr(self):
        if self.isIpfsRoot:
            return shortCidRepr(self._rootCid)

        if self.isIpnsRoot:
            return self.objPath

        if self.isIpfs or self.isIpns:
            basename = self.basename

            if not basename:
                return ''

            if len(basename) > 16:
                return '... {0}'.format(basename[3 * int(len(basename) / 5):])
            elif len(basename) > 24:
                return '... {0}'.format(basename[4 * int(len(basename) / 5):])
            else:
                return basename

    async def resolve(self, ipfsop, noCache=False, timeout=10):
        """
        Resolve this object's path to a CID
        """

        if self.resolvedCid and noCache is False:
            # Return cached value
            return self.resolvedCid

        haveResolve = await ipfsop.hasCommand('resolve')

        try:
            if haveResolve:
                # Use /api/vx/resolve as the preferred resolve strategy

                resolved = await ipfsop.resolve(self.objPath, timeout=timeout)
                if resolved and isIpfsPath(resolved):
                    resolved = stripIpfs(resolved)

                    if cidValid(resolved):
                        self._resolvedCid = resolved
                        return self.resolvedCid

            # Use object stat
            info = StatInfo(await ipfsop.objStat(
                self.objPath, timeout=timeout))

            if info.valid and cidValid(info.cid):
                self._resolvedCid = info.cid
                return self.resolvedCid
        except aioipfs.APIError:
            log.debug('Error resolving {0}'.format(self.objPath))
            return None

    def __str__(self):
        return self.fullPath if self.valid else 'Invalid path: {}'.format(
            self.input)

    def __eq__(self, other):
        if isinstance(other, IPFSPath) and other.valid and self.valid:
            return self.fullPath == other.fullPath
        return False

    def __repr__(self):
        return '{path} (fragment: {frag})'.format(
            path=self.fullPath,
            frag=self.fragment if self.fragment else 'No fragment')


def ipfsRegSearchPath(text):
    return ipfsPathRe.match(text)


def ipfsDedSearchPath(text):
    # search for the dedicated ipfs:// scheme
    return ipfsPathDedRe.match(text)


def ipfsDedSearchPath58(text):
    # Like ipfsDedSearchPath() but allows any type of CID (including
    # CIDv0) as root CID. Only used to be able to extract the root CID
    # and replace it with the base32 version
    return ipfsPathDedRe58.match(text)


def ipfsRegSearchCid(text):
    return ipfsCidRe.match(text)


def ipfsRegSearchCid32(text):
    return ipfsCid32Re.match(text)


def ipnsRegSearchPath(text):
    return ipnsPathRe.match(text) or ipnsPathDedRe.match(text)


def ipfsPathExtract(text):
    ma = ipfsRegSearchPath(text)
    if ma:
        return ma.group('fullpath')

    ma = ipnsRegSearchPath(text)
    if ma:
        return ma.group('fullpath')

    if ipfsRegSearchCid(text):
        return joinIpfs(text)



import functools

from PyQt5.QtCore import Qt
from PyQt5.QtCore import QFile
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QCoreApplication

from PyQt5.QtGui import QImage
from PyQt5.QtGui import QMovie
from PyQt5.QtGui import QPixmap
from PyQt5.QtGui import QPalette
from PyQt5.QtGui import QKeySequence

from PyQt5.Qt import QSizePolicy

from PyQt5.QtWidgets import QFrame
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QScrollArea
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QSpacerItem
from PyQt5.QtWidgets import QLayout
from PyQt5.QtWidgets import QApplication

from galacteek import ensure
from galacteek import logUser
from galacteek.crypto.qrcode import IPFSQrDecoder
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.stat import StatInfo
from galacteek.ipfs.mimetype import MIMEType

from .widgets import GalacteekTab
from .widgets import IPFSUrlLabel
from .widgets import IPFSPathClipboardButton
from .helpers import getIcon
from .helpers import messageBox
from .clipboard import iCopyToClipboard
from .i18n import iZoomIn
from .i18n import iZoomOut
from .i18n import iOpen


def iImageGotQrCodes(count):
    return QCoreApplication.translate(
        'ImageView',
        'This image contains {} valid IPFS QR code(s) !').format(
            count)


def iImageCannotLoad(path):
    return QCoreApplication.translate(
        'ImageView',
        'Cannot load image {}').format(path)


class ImageViewerTab(GalacteekTab):
    def __init__(self, mainW):
        super().__init__(mainW)

        self.pinButton = QToolButton(self)
        self.pinButton.setIcon(getIcon('pin.png'))
        self.pinButton.setEnabled(True)

        self.zoomIn = QToolButton(self)
        self.zoomIn.setIcon(getIcon('zoom-in.png'))
        self.zoomIn.setShortcut(QKeySequence('Ctrl++'))
        self.zoomIn.setToolTip(iZoomIn())
        self.zoomOut = QToolButton(self)
        self.zoomOut.setIcon(getIcon('zoom-out.png'))
        self.zoomOut.setShortcut(QKeySequence('Ctrl+-'))
        self.zoomOut.setToolTip(iZoomOut())

        self.fitWindow = QToolButton(self)
        self.fitWindow.setCheckable(True)
        self.fitWindow.setToolTip('Fit to window')
        self.fitWindow.setIcon(getIcon('expand.png'))

        self.pathLabel = IPFSUrlLabel(None)
        self.pathClipB = IPFSPathClipboardButton(None)

        layout = QHBoxLayout()
        layout.addWidget(self.pathLabel, 0, Qt.AlignLeft)
        layout.addWidget(self.pathClipB, 0, Qt.AlignLeft)
        layout.addItem(
            QSpacerItem(10, 10, QSizePolicy.Expanding, QSizePolicy.Minimum))
        layout.addWidget(self.pinButton, 0, Qt.AlignLeft)
        layout.addItem(
            QSpacerItem(10, 10, QSizePolicy.Minimum, QSizePolicy.Minimum))
        layout.addWidget(self.zoomOut, 0, Qt.AlignLeft)
        layout.addWidget(self.zoomIn, 0, Qt.AlignLeft)
        layout.addWidget(self.fitWindow, 0, Qt.AlignLeft)

        self.vLayout.addLayout(layout)

        self.view = ImageView(self)
        self.view.imageLoaded.connect(self.onImageLoaded)
        self.view.qrCodesPresent.connect(self.onQrCodesListed)
        self.vLayout.addWidget(self.view)

        self.zoomIn.clicked.connect(self.view.zoomInClicked)
        self.zoomOut.clicked.connect(self.view.zoomOutClicked)
        self.fitWindow.clicked.connect(
            lambda: self.view.fitWindow(self.fitWindow.isChecked()))
        self.pinButton.clicked.connect(self.view.pinImage)

    def onImageLoaded(self, path, mimeType):
        self.pinButton.setEnabled(True)
        self.pathLabel.path = path
        self.pathClipB.path = path

    def onQrCodesListed(self, urls):
        # Create the scroll area and fix maximum height
        # TODO: set maximum size on resize as well
        scrollArea = QScrollArea()
        scrollArea.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding)
        scrollArea.setWidgetResizable(True)
        scrollArea.setMaximumHeight(self.height() / 3)

        frame = QFrame(scrollArea)
        frame.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding)
        layout = QVBoxLayout()
        layout.setSizeConstraint(QLayout.SetMinAndMaxSize)

        scrollArea.setWidget(frame)

        iconOpen = getIcon('open.png')
        lbl = QLabel()
        lbl.setText(iImageGotQrCodes(len(urls)))
        lbl.setObjectName('qrCodeCountLabel')
        lbl.setStyleSheet('QLabel { font-size: 14pt; text-align: center; }')
        layout.addWidget(lbl)
        layout.addItem(
            QSpacerItem(10, 10, QSizePolicy.Minimum, QSizePolicy.Minimum))

        def onOpenItem(url):
            ensure(self.app.resourceOpener.open(url))

        for url in urls:
            hLayout = QHBoxLayout()
            hLayout.setSizeConstraint(QLayout.SetMinAndMaxSize)
            urlString = str(url)

            if len(urlString) > 92:
                urlText = urlString[0:92] + '...'
            else:
                urlText = urlString

            lbl = QLabel()
            lbl.setText('<b>{}</b>'.format(urlText))
            lbl.setToolTip(urlString)
            lbl.setStyleSheet('QLabel { font-size: 12pt }')

            clipBtn = QToolButton(self)
            clipBtn.setIcon(getIcon('clipboard.png'))
            clipBtn.setToolTip(iCopyToClipboard())
            clipBtn.clicked.connect(
                functools.partial(self.app.setClipboardText, str(url)))

            openBtn = QToolButton(self)
            openBtn.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
            openBtn.setIcon(iconOpen)
            openBtn.setText(iOpen())
            openBtn.clicked.connect(functools.partial(onOpenItem, url))

            hLayout.addWidget(lbl)
            hLayout.addWidget(openBtn)
            hLayout.addWidget(clipBtn)
            layout.addLayout(hLayout)

        frame.setLayout(layout)
        self.vLayout.addWidget(scrollArea)


class ImageView(QScrollArea):
    imageLoaded = pyqtSignal(IPFSPath, MIMEType)

    qrCodesPresent = pyqtSignal(list)

    def __init__(self, parent=None):
        super(ImageView, self).__init__(parent)

        self.app = QApplication.instance()
        self.image = QImage()

        self.clip = QMovie(self)
        self.clipFirstPixmap = None
        self.clipCurrentPixmap = None

        self.currentImgPath = None
        self.setObjectName('ImageView')

        self.qrDecoder = IPFSQrDecoder()
        self.labelImage = QLabel()
        self.setAlignment(Qt.AlignCenter)
        self.labelImage.setAlignment(Qt.AlignCenter)

        self.setBackgroundRole(QPalette.Dark)
        self.labelImage.setSizePolicy(
            QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.labelImage.setScaledContents(True)
        self.scaleFactor = 1.0
        self.setWidget(self.labelImage)

    def wheelEvent(self, event):
        """
        Zoom in/out on the image on mouse wheel events
        """

        degrees = event.angleDelta()

        if degrees.y() == 0:
            return

        yDeg = max(min(60, abs(degrees.y())), 240)

        if degrees.y() > 0:
            factor = 1 + ((yDeg / 30) * 0.05)
            self.zoomIn(factor=factor)
        elif degrees.y() < 0:
            factor = 1 - ((yDeg / 30) * 0.05)
            self.zoomOut(factor=factor)

        event.accept()

    def pinImage(self):
        if self.currentImgPath and self.currentImgPath.valid:
            ensure(self.pinImageFromPath(self.currentImgPath))

    @ipfsOp
    async def pinImageFromPath(self, ipfsop, imgPath):
        await ipfsop.ctx.pin(str(imgPath))

    def fitWindow(self, fit):
        self.setWidgetResizable(fit)

        if not fit:
            self.scaleFactor = 1.0
            self.resizePixmap()

    def zoomOut(self, factor=0.75):
        self.scaleFactor *= factor
        self.resizePixmap()

    def zoomIn(self, factor=1.25):
        self.scaleFactor *= factor
        self.resizePixmap()

    def zoomInClicked(self):
        self.zoomIn()

    def zoomOutClicked(self):
        self.zoomOut()

    def resizePixmap(self):
        if self.labelImage.movie() and self.clipFirstPixmap:
            self.labelImage.resize(
                self.scaleFactor * self.clipFirstPixmap.size())
            return

        pixmap = self.labelImage.pixmap()

        if pixmap == 0 or pixmap is None:
            return

        self.labelImage.resize(
            self.scaleFactor * pixmap.size())

    @ipfsOp
    async def showImage(self, ipfsop, imgPath, fromBlock=False, timeout=10):
        try:
            if fromBlock is True:
                imgData = await ipfsop.waitFor(
                    ipfsop.client.block.get(imgPath), timeout
                )
            else:
                imgData = await ipfsop.waitFor(
                    ipfsop.client.cat(imgPath), timeout
                )

            if not imgData:
                raise Exception('Failed to load image')

            mimeType, statInfo = await self.app.rscAnalyzer(
                IPFSPath(imgPath))

            if mimeType is None:
                raise Exception('Failed to load image')

            stat = StatInfo(statInfo)

            if mimeType.isAnimation:
                fp = self.app.tempDir.filePath(stat.cid)
                tmpFile = QFile(fp)

                if not tmpFile.exists() and tmpFile.open(QFile.ReadWrite):
                    tmpFile.write(imgData)
                    tmpFile.close()

                self.clip.stop()
                self.clip.setFileName(fp)
                self.labelImage.setMovie(self.clip)
                self.clip.frameChanged.connect(self.onMovieFrameChanged)
                self.clip.error.connect(self.onMovieError)
                self.clip.start()
                self.labelImage.adjustSize()
            else:
                img = QImage()
                img.loadFromData(imgData)
                self.image = img

                self.labelImage.setPixmap(QPixmap.fromImage(self.image))
                self.labelImage.adjustSize()
                self.resizePixmap()

                if self.qrDecoder:
                    # See if we have any QR codes in the image
                    urls = self.qrDecoder.decode(imgData)
                    if urls:
                        self.qrCodesPresent.emit(urls)

            self.labelImage.adjustSize()
            self.currentImgPath = IPFSPath(imgPath)
            self.imageLoaded.emit(self.currentImgPath, mimeType)

        except Exception:
            logUser.debug('Failed to load image: {path}'.format(
                path=imgPath))
            messageBox(iImageCannotLoad(imgPath))

    def onMovieFrameChanged(self, num):
        self.clipCurrentPixmap = self.clip.currentPixmap()

        if self.clipFirstPixmap is None:
            self.clipFirstPixmap = self.clipCurrentPixmap

    def onMovieError(self, err):
        messageBox(self.clip.lastErrorString())




class StatInfo:
    def __init__(self, statInfo):
        self.stat = statInfo

    @property
    def valid(self):
        # :-/
        return isinstance(self.stat, dict) and \
            'DataSize' in self.stat and \
            'CumulativeSize' in self.stat and \
            'NumLinks' in self.stat and \
            'BlockSize' in self.stat and \
            'LinksSize' in self.stat

    @property
    def cid(self):
        return self.stat.get('Hash')

    @property
    def totalSize(self):
        return self.stat.get('CumulativeSize')

    @property
    def dataSize(self):
        return self.stat.get('DataSize')

    @property
    def numLinks(self):
        return self.stat.get('NumLinks')

    @property
    def blockSize(self):
        return self.stat.get('BlockSize')

    @property
    def linksSize(self):
        return self.stat.get('LinksSize')

    def dataLargerThan(self, size):
        if self.valid:
            return self.dataSize > size

    def dataSmallerThan(self, size):
        if self.valid:
            return self.dataSize < size



from urllib.parse import quote

from galacteek import log

import aiohttp

ipfsSearchApiHost = 'api.ipfs-search.com'


class IPFSSearchResults:
    def __init__(self, page, results):
        self.pageCount = results.get('page_count', 0)
        self.resultsCount = results.get('total', 0)
        self.page = page
        self.results = results

    @property
    def hits(self):
        return self.results.get('hits', [])

    @property
    def hitsCount(self):
        return len(self.hits)

    def findByHash(self, hashV):
        for hit in self.hits:
            hitHash = hit.get('hash', None)
            if hitHash == hashV:
                return hit


emptyResults = IPFSSearchResults(0, {})


async def searchPage(query, page, filters={}, sslverify=True):
    params = {
        'q': query,
        'page': page
    }

    for fkey, fvalue in filters.items():
        params['q'] += quote(' {fkey}:{fvalue}'.format(
            fkey=fkey, fvalue=fvalue))

    async with aiohttp.ClientSession() as session:
        async with session.get('https://{host}/v1/search'.format(
                host=ipfsSearchApiHost),
                params=params,
                verify_ssl=sslverify) as resp:
            return await resp.json()


async def getPageResults(query, page, filters={}, sslverify=True):
    try:
        results = await searchPage(query, page, filters=filters,
                                   sslverify=sslverify)
        return IPFSSearchResults(page, results)
    except Exception:
        return None


async def getMetadata(cid, sslverify=True):
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get('https://{host}/v1/metadata/{cid}'.format(
                    host=ipfsSearchApiHost, cid=cid),
                    verify_ssl=sslverify) as resp:
                return await resp.json()
    except Exception:
        log.debug('Error occured while fetching metadata')
        return None


async def search(query, pageStart=0, preloadPages=0,
                 filters={}, sslverify=True):
    page1Results = await getPageResults(query, pageStart, filters=filters,
                                        sslverify=sslverify)
    if page1Results is None:
        return

    yield page1Results
    pageCount = page1Results.pageCount

    if preloadPages > 0:
        pageLast = preloadPages + pageStart if \
            pageCount >= preloadPages else pageCount

        for page in range(page1Results.page + 1, pageLast):
            results = await getPageResults(query, page, filters=filters,
                                           sslverify=sslverify)
            if results:
                yield results



import re
import aioipfs

from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QComboBox
from PyQt5.QtWidgets import QInputDialog
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QDialogButtonBox
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QGridLayout
from PyQt5.QtWidgets import QFormLayout

from PyQt5.QtCore import QSize
from PyQt5.QtCore import QFile
from PyQt5.QtCore import QIODevice
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import QRegExp
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QUrl

from PyQt5.QtGui import QClipboard
from PyQt5.QtGui import QPixmap
from PyQt5.QtGui import QImage
from PyQt5.QtGui import QRegExpValidator

from galacteek import GALACTEEK_NAME
from galacteek import asyncify
from galacteek import ensure
from galacteek import logUser

from galacteek.core.ipfsmarks import *
from galacteek.core import countries
from galacteek.core.ipfsmarks import categoryValid
from galacteek.core.profile import UserInfos
from galacteek.ipfs import cidhelpers
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.wrappers import ipfsOp, ipfsOpFn

from . import ui_addhashmarkdialog
from . import ui_addfeeddialog
from . import ui_ipfscidinputdialog, ui_ipfsmultiplecidinputdialog
from . import ui_profileeditdialog
from . import ui_donatedialog
from . import ui_qschemecreatemapping

from .helpers import *
from .widgets import ImageWidget
from .widgets import HorizontalLine
from .widgets import IconSelector
from .widgets import LabelWithURLOpener

from .i18n import iTitle
from .i18n import iDoNotPin
from .i18n import iPinSingle
from .i18n import iPinRecursive
from .i18n import iNoTitleProvided


def boldLabelStyle():
    return 'QLabel { font-weight: bold; }'


class AddHashmarkDialog(QDialog):
    def __init__(
            self,
            marks,
            resource,
            title,
            description,
            stats,
            pin=False,
            pinRecursive=False,
            parent=None):
        super().__init__(parent)

        self.ipfsResource = resource
        self.marks = marks
        self.stats = stats if stats else {}
        self.iconCid = None

        self.ui = ui_addhashmarkdialog.Ui_AddHashmarkDialog()
        self.ui.setupUi(self)
        self.ui.resourceLabel.setText(self.ipfsResource)
        self.ui.resourceLabel.setStyleSheet(boldLabelStyle())
        self.ui.resourceLabel.setToolTip(self.ipfsResource)
        self.ui.newCategory.textChanged.connect(self.onNewCatChanged)
        self.ui.title.setText(title)

        pix = QPixmap.fromImage(QImage(':/share/icons/hashmarks.png'))
        pix = pix.scaledToWidth(32)
        self.ui.hashmarksIconLabel.setPixmap(pix)

        self.iconWidget = None

        self.ui.pinCombo.addItem(iDoNotPin())
        self.ui.pinCombo.addItem(iPinSingle())
        self.ui.pinCombo.addItem(iPinRecursive())

        self.ui.formLayout.setRowWrapPolicy(QFormLayout.DontWrapRows)
        self.ui.formLayout.setFieldGrowthPolicy(
            QFormLayout.ExpandingFieldsGrow)
        self.ui.formLayout.setLabelAlignment(Qt.AlignHCenter | Qt.AlignTop)
        self.ui.formLayout.setHorizontalSpacing(20)

        self.iconSelector = IconSelector(parent=self, allowEmpty=True)
        self.iconSelector.iconSelected.connect(self.onIconSelected)
        self.iconSelector.emptyIconSelected.connect(self.onIconEmpty)
        self.ui.formLayout.insertRow(7, QLabel('Icon'),
                                     self.iconSelector)

        regexp1 = QRegExp(r"[A-Za-z0-9/\-]+")  # noqa
        self.ui.newCategory.setValidator(QRegExpValidator(regexp1))
        self.ui.newCategory.setMaxLength(64)

        if pin is True:
            self.ui.pinCombo.setCurrentIndex(1)
        elif pinRecursive is True:
            self.ui.pinCombo.setCurrentIndex(2)

        if isinstance(description, str):
            self.ui.description.insertPlainText(description)

        for cat in self.marks.getCategories():
            self.ui.category.addItem(cat)

    def onIconSelected(self, iconCid):
        self.iconCid = iconCid

    def onIconEmpty(self):
        self.iconCid = None

    def onSelectIcon(self):
        fps = filesSelectImages()
        if len(fps) > 0:
            ensure(self.setIcon(fps.pop()))

    @ipfsOp
    async def setIcon(self, op, fp):
        entry = await op.addPath(fp, recursive=False)
        if entry:
            cid = entry['Hash']

            if self.iconWidget is None:
                iconWidget = ImageWidget()

                if await iconWidget.load(cid):
                    self.ui.formLayout.insertRow(7, QLabel(''), iconWidget)
                    self.iconCid = cid
                    self.iconWidget = iconWidget
            else:
                if await self.iconWidget.load(cid):
                    self.iconCid = cid

    def onNewCatChanged(self, text):
        self.ui.category.setEnabled(len(text) == 0)

    def accept(self):
        title = self.ui.title.text()

        if len(title) == 0:
            return messageBox(iNoTitleProvided())

        share = self.ui.share.isChecked()
        newCat = self.ui.newCategory.text()
        description = self.ui.description.toPlainText()

        if len(description) > 1024:
            return messageBox('Description is too long')

        if len(newCat) > 0:
            category = cidhelpers.normp(newCat)
        else:
            category = self.ui.category.currentText()

        pSingle = (self.ui.pinCombo.currentIndex() == 1)
        pRecursive = (self.ui.pinCombo.currentIndex() == 2)

        mark = IPFSHashMark.make(
            self.ipfsResource,
            title=title,
            share=share,
            comment=self.ui.comment.text(),
            description=description,
            pinSingle=pSingle,
            pinRecursive=pRecursive,
            icon=self.iconCid,
            datasize=self.stats.get(
                'DataSize',
                None),
            cumulativesize=self.stats.get(
                'CumulativeSize',
                None),
            numlinks=self.stats.get(
                'NumLinks',
                None))

        self.marks.insertMark(mark, category)
        self.done(0)


class AddFeedDialog(QDialog):
    def __init__(self, marks, resource, feedName=None, parent=None):
        super().__init__(parent)

        self.marks = marks
        self.ipfsResource = resource

        self.ui = ui_addfeeddialog.Ui_AddFeedDialog()
        self.ui.setupUi(self)
        self.ui.resourceLabel.setText(self.ipfsResource)
        self.ui.resourceLabel.setStyleSheet(boldLabelStyle())

        self.ui.formLayout.setRowWrapPolicy(QFormLayout.DontWrapRows)
        self.ui.formLayout.setFieldGrowthPolicy(
            QFormLayout.ExpandingFieldsGrow)
        self.ui.formLayout.setLabelAlignment(Qt.AlignCenter)
        self.ui.formLayout.setHorizontalSpacing(20)

        if isinstance(feedName, str):
            self.ui.feedName.setText(feedName)

    def accept(self):
        share = self.ui.share.isChecked()
        autoPin = self.ui.autoPin.isChecked()
        feedName = self.ui.feedName.text()

        if len(feedName) == 0:
            return messageBox('Please specify a feed name')

        self.marks.follow(self.ipfsResource, self.ui.feedName.text(),
                          resolveevery=self.ui.resolve.value(),
                          share=share, autoPin=autoPin)
        self.done(0)


class IPFSCIDInputDialog(QDialog):
    """ Dialog for IPFS CID input and validation """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.validCid = False
        self.ui = ui_ipfscidinputdialog.Ui_CIDInputDialog()
        self.ui.setupUi(self)
        self.ui.validity.setStyleSheet(boldLabelStyle())
        self.ui.cid.textChanged.connect(self.onCIDChanged)
        self.ui.clearButton.clicked.connect(self.onClearCID)

    def onClearCID(self):
        self.clearCID()

    def clearCID(self):
        """ Clears the CID input text """
        self.ui.cid.clear()

    def onCIDChanged(self, text):
        """
        When the CID input text changes we verify if it's a valid CID and
        update the validity label
        """
        if cidhelpers.cidValid(text):
            cid = self.getCID()
            if cid:
                self.ui.validity.setText(
                    'Valid CID version {}'.format(cid.version))
                self.validCid = True
            else:
                # Unlikely
                self.ui.validity.setText('Unknown CID type')
                self.validCid = False
        else:
            self.ui.validity.setText('Invalid CID')
            self.validCid = False

    def getHash(self):
        """ Returns the hash corresponding to the input, if valid """
        cid = self.getCID()
        if cid:
            return str(cid)

    def getCID(self):
        """ Returns the CID object corresponding to the input """
        try:
            return cidhelpers.getCID(self.ui.cid.text())
        except BaseException:
            return None

    def accept(self):
        if self.validCid is True:
            self.done(1)


class IPFSMultipleCIDInputDialog(QDialog):
    """ Dialog for multiple IPFS CID input """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.validCid = False
        self.ui = ui_ipfsmultiplecidinputdialog.Ui_MultipleCIDInputDialog()
        self.ui.setupUi(self)
        self.ui.validity.setStyleSheet(boldLabelStyle())
        self.ui.cid.textChanged.connect(self.onCIDChanged)
        self.ui.addCidButton.clicked.connect(self.onAddCID)

    def onAddCID(self):
        if self.validCid:
            self.ui.cidList.addItem(self.ui.cid.text())
            self.ui.cid.clear()

    def onCIDChanged(self, text):
        if cidhelpers.cidValid(text):
            self.ui.validity.setText('Valid CID')
            self.validCid = True
        else:
            self.ui.validity.setText('Invalid CID')
            self.validCid = False

    def getCIDs(self):
        cids = []
        for idx in range(0, self.ui.cidList.count()):
            item = self.ui.cidList.item(idx)
            cids.append(item.text())
        return cids

    def accept(self):
        self.done(1)


class DonateDialog(QDialog):
    def __init__(self, bcAddr, parent=None):
        super().__init__(parent)

        self.app = QApplication.instance()
        self.ui = ui_donatedialog.Ui_DonateDialog()
        self.ui.setupUi(self)
        self.ui.okButton.clicked.connect(self.close)
        self.ui.bitcoinClip.clicked.connect(self.onBcClip)
        self.ui.donatePatreon.clicked.connect(self.onPatreonClicked)

        self._bcAddr = bcAddr
        self.ui.bitcoinAddress.setText('<b>{0}</b>'.format(self._bcAddr))
        self.setWindowTitle('Make a donation')

    def onPatreonClicked(self):
        tab = self.app.mainWindow.addBrowserTab()
        tab.enterUrl(QUrl(
            'https://www.patreon.com/{0}'.format(GALACTEEK_NAME)))
        self.accept()

    def onBcClip(self):
        self.toClip(self._bcAddr)

    def toClip(self, data):
        app = QCoreApplication.instance()
        app.clipboard().setText(data, QClipboard.Selection)
        app.clipboard().setText(data, QClipboard.Clipboard)

    def close(self):
        self.done(1)

    def accept(self):
        self.done(1)


def notEmpty(v):
    return v != ''


class ProfileEditDialog(QDialog):
    genderMapping = {
        'Unspecified': UserInfos.GENDER_UNSPECIFIED,
        'Male': UserInfos.GENDER_MALE,
        'Female': UserInfos.GENDER_FEMALE
    }

    def __init__(self, profile, parent=None):
        super().__init__(parent)

        self.profile = profile
        self.countryList = countries.countryList
        self.previousUsername = self.profile.userInfo.username

        self.ui = ui_profileeditdialog.Ui_ProfileEditDialog()
        self.ui.setupUi(self)
        self.ui.tabWidget.setCurrentIndex(0)

        self.loadCountryData()

        self.ui.labelWarning.setStyleSheet('QLabel { font-weight: bold; }')
        self.ui.username.setText(self.profile.userInfo.username)
        self.ui.firstname.setText(self.profile.userInfo.firstname)
        self.ui.lastname.setText(self.profile.userInfo.lastname)
        self.ui.email.setText(self.profile.userInfo.email)
        self.ui.org.setText(self.profile.userInfo.org)
        self.ui.city.setText(self.profile.userInfo.city)
        self.ui.bio.setText(self.profile.userInfo.bio)

        for gender, gvalue in self.genderMapping.items():
            self.ui.gender.addItem(gender)
            if self.profile.userInfo.gender == gvalue:
                self.ui.gender.setCurrentText(gender)

        if notEmpty(self.profile.userInfo.countryName):
            self.ui.countryBox.setCurrentText(
                self.profile.userInfo.countryName)
        else:
            self.ui.countryBox.setCurrentText('Unspecified')

        if notEmpty(self.profile.userInfo.avatarCid):
            self.updateAvatarCid()

        self.ui.profileCryptoId.setText('<b>{0}</b>'.format(
            self.profile.userInfo.objHash))

        self.ui.changeIconButton.clicked.connect(self.changeIcon)
        self.ui.updateButton.clicked.connect(self.save)
        self.ui.cancelButton.clicked.connect(self.close)

        self.reloadIcon()

    def updateAvatarCid(self):
        self.ui.iconHash.setText('<a href="ipfs:{0}">{1}</a>'.format(
            cidhelpers.joinIpfs(self.profile.userInfo.avatarCid),
            self.profile.userInfo.avatarCid))

    def reloadIcon(self):
        ensure(self.loadIcon())

    def getCountryCode(self, name):
        for entry in self.countryList:
            if entry['name'] == name:
                return entry['code']

    def loadCountryData(self):
        for entry in self.countryList:
            self.ui.countryBox.addItem(entry['name'])

        self.ui.countryBox.addItem('Unspecified')

    @asyncify
    async def loadIcon(self):
        @ipfsOpFn
        async def load(op, cid):
            try:
                imgData = await op.client.cat(cid)
                img1 = QImage()
                img1.loadFromData(imgData)
                img = img1.scaledToWidth(256)
                self.ui.iconPixmap.setPixmap(QPixmap.fromImage(img))
            except Exception:
                messageBox('Error while loading image')

        if self.profile.userInfo.avatarCid != '':
            await load(self.profile.userInfo.avatarCid)

    def changeIcon(self):
        fps = filesSelectImages()
        if len(fps) > 0:
            ensure(self.setIcon(fps.pop()))

    @ipfsOp
    async def setIcon(self, op, fp):
        entry = await op.addPath(fp, recursive=False)
        if entry:
            self.profile.userInfo.setAvatarCid(entry['Hash'])
            self.reloadIcon()
            self.updateAvatarCid()

    def save(self):
        kw = {}
        for key in ['username', 'firstname',
                    'lastname', 'email', 'org',
                    'city']:
            val = getattr(self.ui, key).text()
            ma = re.search(r"[a-zA-Z\_\-\@\.\+\'\´0-9\s]*", val)
            if ma:
                kw[key] = val

        try:
            kw['bio'] = self.ui.bio.toPlainText()
        except:
            pass

        country = self.ui.countryBox.currentText()
        code = self.getCountryCode(country)

        if country and code:
            self.profile.userInfo.setCountryInfo(country, code)

        genderSelected = self.ui.gender.currentText()
        for gender, gvalue in self.genderMapping.items():
            if genderSelected == gender:
                kw['gender'] = gvalue

        self.profile.userInfo.setInfos(**kw)

        if self.previousUsername != self.profile.userInfo.username:
            self.profile.userInfo.usernameChanged.emit()

        self.profile.userInfo.changed.emit()
        self.done(1)

    def reject(self):
        self.done(0)


class ChooseProgramDialog(QInputDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle('Choose a program')
        self.setInputMode(QInputDialog.TextInput)
        self.setLabelText(
            '''Command arguments (example: <b>mupdf %f</b>).
                <b>%f</b> is replaced with the file path''')


class AddAtomFeedDialog(QInputDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle('Add an Atom feed')
        self.setInputMode(QInputDialog.TextInput)
        self.setLabelText('Atom feed URL')

    def sizeHint(self):
        return QSize(600, 140)


class AddMultihashPyramidDialog(QDialog):
    def __init__(self, marks, pyramidType, parent=None):
        super().__init__(parent)

        self.app = QApplication.instance()

        self.marks = marks
        self.pyramidType = pyramidType
        self.iconCid = None
        self.customCategory = None

        cat = QLabel('Category')
        self.categoryCombo = QComboBox(self)
        cat.setBuddy(self.categoryCombo)

        for category in marks.getCategories():
            self.categoryCombo.addItem(category)

        label = QLabel('Pyramid name')
        restrictRegexp = QRegExp("[0-9A-Za-z-_]+")  # noqa
        restrictRegexp2 = QRegExp("[0-9A-Za-z-_/]+")  # noqa

        self.nameLine = QLineEdit()
        self.nameLine.setMaxLength(32)
        self.nameLine.setValidator(QRegExpValidator(restrictRegexp))
        label.setBuddy(self.nameLine)

        buttonBox = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel, self)

        catLayout = QHBoxLayout()
        catLayout.addWidget(cat)
        catLayout.addWidget(self.categoryCombo)

        catCustomLabel = QLabel('Or create new category')
        catCustomLayout = QHBoxLayout()

        self.catCustom = QLineEdit()
        self.catCustom.setMaxLength(32)
        self.catCustom.setValidator(QRegExpValidator(restrictRegexp2))
        self.catCustom.textChanged.connect(self.onCustomCategory)
        catCustomLabel.setBuddy(self.catCustom)
        catCustomLayout.addWidget(catCustomLabel)
        catCustomLayout.addWidget(self.catCustom)

        descrLabel = QLabel('Description')
        self.descrLine = QLineEdit()
        self.descrLine.setMaxLength(128)
        descrLabel.setBuddy(self.descrLine)
        descrLayout = QHBoxLayout()
        descrLayout.addWidget(descrLabel)
        descrLayout.addWidget(self.descrLine)

        nameLayout = QHBoxLayout()
        nameLayout.addWidget(label)
        nameLayout.addWidget(self.nameLine)

        self.iconSelector = IconSelector()
        self.iconSelector.iconSelected.connect(self.onIconSelected)

        self.lifetimeCombo = QComboBox(self)
        self.lifetimeCombo.addItem('24h')
        self.lifetimeCombo.addItem('48h')
        self.lifetimeCombo.addItem('96h')
        self.lifetimeCombo.setCurrentText('48h')
        ipnsLTimeLayout = QHBoxLayout()
        ipnsLTimeLayout.addWidget(QLabel('IPNS record lifetime'))
        ipnsLTimeLayout.addWidget(self.lifetimeCombo)

        pickIconLayout = QHBoxLayout()
        pickIconLayout.addWidget(QLabel('Choose icon'))
        pickIconLayout.addWidget(self.iconSelector)

        mainLayout = QGridLayout()
        mainLayout.addLayout(nameLayout, 0, 0)
        mainLayout.addLayout(descrLayout, 1, 0)
        mainLayout.addWidget(HorizontalLine(self), 2, 0)
        mainLayout.addLayout(catLayout, 3, 0)
        mainLayout.addLayout(catCustomLayout, 4, 0)
        mainLayout.addWidget(HorizontalLine(self), 5, 0)
        mainLayout.addLayout(ipnsLTimeLayout, 6, 0)
        mainLayout.addLayout(pickIconLayout, 7, 0)
        mainLayout.addWidget(buttonBox, 8, 0)

        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        self.setLayout(mainLayout)

        self.nameLine.setFocus(Qt.OtherFocusReason)

    def onIconSelected(self, iconCid):
        self.iconCid = iconCid

    def onCustomCategory(self, text):
        self.categoryCombo.setEnabled(len(text) == 0)
        self.customCategory = text

    def reject(self):
        self.done(0)

    def accept(self):
        pyramidName = self.nameLine.text()
        descr = self.descrLine.text()
        lifetime = self.lifetimeCombo.currentText()

        if len(pyramidName) == 0 or len(descr) == 0:
            return messageBox('Please give a name and description')

        if isinstance(self.customCategory, str) and \
                categoryValid(self.customCategory):
            category = self.customCategory
        else:
            category = self.categoryCombo.currentText()

        ipnsKeyName = 'galacteek.pyramids.{cat}.{name}'.format(
            cat=category.replace('/', '_'), name=pyramidName)

        self.done(1)

        ensure(self.createPyramid(pyramidName, category, ipnsKeyName,
                                  descr, lifetime))

    @ipfsOp
    async def createPyramid(self, ipfsop, pyramidName, category, ipnsKeyName,
                            description, ipnsLifetime):
        try:
            logUser.info(
                'Multihash pyramid {pyr}: generating IPNS key ...'.format(
                    pyr=pyramidName))
            ipnsKey = await ipfsop.keyGen(ipnsKeyName)
        except aioipfs.APIError:
            return
        else:
            if ipnsKey:
                self.marks.pyramidNew(
                    pyramidName, category, self.iconCid,
                    ipnskey=ipnsKey['Id'],
                    lifetime=ipnsLifetime,
                    type=self.pyramidType,
                    description=description)
                logUser.info('Multihash pyramid {pyr}: created'.format(
                    pyr=pyramidName))

    @ipfsOp
    async def injectQrcIcon(self, op, iconPath):
        _file = QFile(iconPath)
        if _file.exists():
            if not _file.open(QIODevice.ReadOnly):
                return False

            data = _file.readAll().data()
            entry = await op.addBytes(data, offline=True)
            if entry:
                self.iconCid = entry['Hash']
                return True


class AboutDialog(QDialog):
    def __init__(self, text, parent=None):
        super().__init__(parent)

        buttonBox = QDialogButtonBox(
            QDialogButtonBox.Ok, self)
        buttonBox.accepted.connect(self.accept)

        self.setWindowTitle('About')

        layout = QVBoxLayout()
        layout.addWidget(LabelWithURLOpener(text, parent=self))
        layout.addWidget(buttonBox)

        self.setLayout(layout)

    def accept(self):
        self.done(1)


class QSchemeCreateMappingDialog(QDialog):
    def __init__(self, mappedPath, parent=None):
        super().__init__(parent)

        self.app = QApplication.instance()
        self.mappedTo = mappedPath

        self.ui = ui_qschemecreatemapping.Ui_QSchemeMappingDialog()
        self.ui.setupUi(self)
        self.ui.buttonBox.accepted.connect(self.accept)
        self.ui.buttonBox.rejected.connect(self.reject)
        self.ui.mappedPath.setText(
            '<b>{}</b>'.format(str(self.mappedTo)))

        regexp = QRegExp(r"[a-z0-9]+")
        self.ui.mappingName.setValidator(QRegExpValidator(regexp))

    def accept(self):
        name = self.ui.mappingName.text()
        if not name:
            return messageBox('Please provide a mapping name')

        if self.app.marksLocal.qaMap(
                name,
                self.mappedTo,
                ipnsResolveFrequency=self.ui.ipnsResolveFrequency.value()):
            self.app.towers['schemes'].qMappingsChanged.emit()
            self.done(1)
            messageBox(
                'You can now use the quick-access URL '
                '<b>q://{name}</b>'.format(name=name)
            )
        else:
            messageBox(
                'An error ocurred, check that a mapping does not '
                'already exist with that name')


class ResourceOpenConfirmDialog(QDialog):
    message = '''
    <p>IPFS object with path <b>{p}</b> (Content type: <b>{mtype}</b>)
    could not be opened.</p>

    <p>
    Open with system's default application ?
    </p>
    '''

    def __init__(self, rscPath, mType, secureEnv, parent=None):
        super().__init__(parent)

        self.rscPath = rscPath
        self.setWindowTitle('Open resource')

        buttonBox = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
            self
        )
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        layout = QVBoxLayout()

        label = QLabel(self.message.format(
            p=str(rscPath), mtype=mType.type))
        label.setMaximumWidth(600)
        label.setWordWrap(True)

        layout.addWidget(label)

        okButton = buttonBox.button(QDialogButtonBox.Ok)

        if not secureEnv:
            okButton.setEnabled(False)
            ignButton = QToolButton(self)
            ignButton.setCheckable(True)
            ignButton.setText('Let me open it anyway')
            ignButton.toggled.connect(
                lambda state: okButton.setEnabled(state))

            wLabel = QLabel(
                '<p style="color: red"> '
                'You are trying to open this object '
                'from an insecure context.'
                '</font>')
            layout.addWidget(wLabel)
            layout.addWidget(ignButton)

        layout.addWidget(buttonBox)

        self.setLayout(layout)

    def accept(self):
        self.done(1)


class TitleInputDialog(QDialog):
    def __init__(self, title, maxLength=64, parent=None):
        super().__init__(parent)

        self.app = QApplication.instance()
        self.title = title

        self.setWindowTitle(iTitle())

        buttonBox = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
            self
        )
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        layout = QVBoxLayout()
        layout.addWidget(QLabel(iTitle()))

        self.tEdit = QLineEdit(self)
        self.tEdit.setMaxLength(maxLength)
        self.tEdit.setText(title[0:maxLength - 1])

        layout.addWidget(self.tEdit)
        layout.addWidget(buttonBox)

        self.setLayout(layout)

    def sizeHint(self):
        return QSize(
            self.app.desktopGeometry.width() / 2,
            100
        )

    def accept(self):
        self.done(1)



import json
import feedparser
from dateutil import parser

import aioipfs

from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal

from galacteek import log
from galacteek.ipfs import ipfsOp
from galacteek.ipfs.cidhelpers import IPFSPath


DWEB_ATOM_FEEDFN = 'dfeed.atom'


class AtomFeedExistsError(Exception):
    pass


class AtomFeedError(Exception):
    pass


class AtomParseFeedError(Exception):
    pass


def parseFeed(data):
    try:
        feed = feedparser.parse(data)
    except:
        log.debug('Feed parsing error')
        return None
    else:
        return feed


class IPFSAtomFeedEntry(QObject):
    statusChanged = pyqtSignal(int)

    ENTRY_STATUS_NEW = 0
    ENTRY_STATUS_READ = 1

    def __init__(self, feed, entry, status=0):
        super().__init__()
        self._feed = feed
        self._entry = entry
        self._status = status

        self.srow_id = None

    @property
    def feed(self):
        return self._entry

    @property
    def entry(self):
        return self._entry

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, v):
        if isinstance(v, int):
            self._status = v
            self.statusChanged.emit(v)

    @property
    def id(self):
        return self.entry['id']

    @property
    def link(self):
        return self.entry['link']

    @property
    def title(self):
        return self.entry['title']

    @property
    def author(self):
        return self.entry['author']

    @property
    def updatedString(self):
        return self.entry['updated']

    @property
    def updated(self):
        return parser.parse(self.updatedString)

    @property
    def publishedString(self):
        return self.entry['published']

    @property
    def published(self):
        return parser.parse(self.publishedString)

    @property
    def tags(self):
        return [t['term'] for t in self.entry['tags']]

    def markAsRead(self):
        self.status = self.ENTRY_STATUS_READ


class IPFSAtomFeed(QObject):
    """
    An Atom feed from the dweb
    """

    statusForEntryChanged = pyqtSignal(IPFSAtomFeedEntry)

    def __init__(self, feed):
        super().__init__()
        self._feed = feed
        self._entries = []
        self._init()

    def _init(self):
        if 'entries' in self.feed:
            self._entries = [
                IPFSAtomFeedEntry(self, entry) for entry in
                self.feed['entries']
            ]

    @property
    def feed(self):
        return self._feed

    @property
    def feedDump(self):
        return json.dumps(self.feed, indent=4)

    @property
    def feedInfos(self):
        return self.feed['feed']

    @property
    def entries(self):
        return self._entries

    @property
    def title(self):
        return self.feedInfos['title']

    @property
    def language(self):
        return self.feedInfos['language']

    @property
    def id(self):
        return self.feedInfos['id']

    @property
    def author(self):
        return self.feedInfos['author']

    @property
    def generator(self):
        return self.feedInfos['generator']

    @property
    def updatedString(self):
        return self.feedInfos['updated']

    @property
    def updated(self):
        dateString = self.updatedString
        dt = parser.parse(dateString)
        return dt

    async def read(self):
        for entry in self.entries:
            yield IPFSAtomFeedEntry(entry)


class IPFSAtomFeedParser:
    @ipfsOp
    async def parse(self, ipfsop, path):
        """
        The caller already knows this object contains a feed (no MIME
        detection here)

        path can be an IPFSPath or a string

            dweb:/ipns/blog.ipfs.io/feed.atom
            dweb:/ipns/QmWmtbVRezi6AtFC8BPL1SoPFDT5jobbFpggpAb7xZTAmB/f.atom

        :param str path: feed's object path
        """

        if isinstance(path, IPFSPath):
            objPath = path.objPath
        elif isinstance(path, str):
            objPath = path
        else:
            log.debug('Unknown obj type to parse: {0}'.format(path))
            return

        try:
            data = await ipfsop.waitFor(
                ipfsop.client.cat(objPath), 10
            )

            if not data:
                return None

            feed = parseFeed(data.decode())
            if feed:
                return IPFSAtomFeed(feed=feed)
            else:
                raise AtomParseFeedError()
        except aioipfs.APIError:
            raise AtomParseFeedError()
        except Exception:
            raise AtomParseFeedError()



import aioipfs

from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QObject

from galacteek import log

from galacteek.ipfs import ipfsOp

from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.mimetype import MIMEType
from galacteek.ipfs.mimetype import detectMimeType

from galacteek.crypto.qrcode import IPFSQrDecoder


class ResourceAnalyzer(QObject):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.app = QApplication.instance()

    @ipfsOp
    async def __call__(self, ipfsop, pathRef):
        """
        :param IPFSPath ipfsPath
        """

        if isinstance(pathRef, IPFSPath):
            ipfsPath = pathRef
        elif isinstance(pathRef, str):
            ipfsPath = IPFSPath(pathRef, autoCidConv=True)
        else:
            raise ValueError('Invalid input value')

        path = ipfsPath.objPath
        mHashMeta = await self.app.multihashDb.get(path)

        if mHashMeta:
            # Already have metadata for this object
            typeStr = mHashMeta.get('mimetype')
            mimetype = MIMEType(typeStr) if typeStr else None
            statInfo = mHashMeta.get('stat')
            return mimetype, statInfo
        else:
            mimetype = await detectMimeType(path)

            statInfo = await ipfsop.objStat(path)
            if not statInfo or not isinstance(statInfo, dict):
                log.debug('Stat failed for {path}'.format(
                    path=path))
                return mimetype, None

            await ipfsop.sleep()

            # Store retrieved information in the metadata store
            metaMtype = mimetype.type if mimetype and mimetype.valid else None
            await self.app.multihashDb.store(
                path,
                mimetype=metaMtype,
                stat=statInfo
            )

        if mimetype and mimetype.valid:
            return mimetype, statInfo

    @ipfsOp
    async def decodeQrCodes(self, ipfsop, path):
        try:
            data = await ipfsop.waitFor(
                ipfsop.client.cat(path), 12
            )

            if data is None:
                return

            # Decode the QR codes in the image if there's any
            qrDecoder = IPFSQrDecoder()
            if not qrDecoder:
                return

            return qrDecoder.decode(data)
        except aioipfs.APIError:
            pass
